"""
Unit and regression test for the wpspec package.
"""

# Import package, test suite, and other packages as needed
import wptherml
import numpy as np
import pytest
import sys
from numpy import linalg
import json
import os

path_and_file = os.path.realpath(__file__)
path = path_and_file[:-23] + "data/json_data/"


# high-level tests for spin-boson model and dynamics
def test_compute_spectrum():

    # test example vs QuTip J-C model for weak coupling between 1 spin and 2 level cavity
    # from json file
    test_1_json = path + "JC_simulation_RWA=_True-spin_freq_0.5_cavity_freq_0.5_cavity_coupling_0.0.json"
    test_2_json = path + "JC_simulation_RWA=_True-spin_freq_0.5_cavity_freq_0.5_cavity_coupling_0.02.json"

    # load json data
    with open(test_1_json) as f:
        json_1_data = json.load(f)

    with open(test_2_json) as f:
        json_2_data = json.load(f)

    

    # test example vs QuTip J-C model for strong coupling between 1 spin and 2 level cavity
    # with spin = cavity omega = 0.5 atomic units, g = 0.02 (using RWA)
    test_args_1 = {
     "Number_of_Excitons": 1,
     "number_of_boson_levels": json_1_data["number_of_cavity_states"],
     "boson_energy_ev": json_1_data["cavity_frequency"] / 3.6749322175665e-2, 
     "exciton_energy_ev" : json_1_data["spin_frequency"] / 3.6749322175665e-2, 
     "exciton_boson_coupling_ev" : json_1_data["cavity_coupling"] / 3.6749322175665e-2
     
    }

    test_args_2 = {
        "Number_of_Excitons": 1,
        "number_of_boson_levels": json_2_data["number_of_cavity_states"],
        "boson_energy_ev": json_2_data["cavity_frequency"] / 3.6749322175665e-2, 
        "exciton_energy_ev" : json_2_data["spin_frequency"] / 3.6749322175665e-2, 
        "exciton_boson_coupling_ev" : json_2_data["cavity_coupling"] / 3.6749322175665e-2
        
        }
    
    test_args_3 = {
     "Number_of_Excitons": 1,
     "number_of_boson_levels": 10,
     "boson_energy_ev": 0.5 / 3.6749322175665e-2, 
     "exciton_energy_ev" : 0.5 / 3.6749322175665e-2, 
     "exciton_boson_coupling_ev" : 0.49 / 3.6749322175665e-2
     
    }

    sf = wptherml.SpectrumFactory()
    
    # instantiate the three tests
    test_1 = sf.spectrum_factory("Spin-Boson", test_args_1)
    test_2 = sf.spectrum_factory("Spin-Boson", test_args_2)
    test_3 = sf.spectrum_factory("Spin-Boson", test_args_3)
    
    # get expected results for the three cases
    _expected_eigs_1 = np.array(json_1_data["energies"])
    _expected_eigs_2 = np.array(json_2_data["energies"])
    _expected_eigs_3 = np.array([0.0, 
                                 0.010000000000000009, 
                                 0.30703535443718355, 
                                 0.6512951042912501, 
                                 0.99, 
                                 1.02, 
                                 1.4043266910251035, 
                                 1.6929646455628165, 
                                 1.799750026036243, 
                                 2.2035818575783503, 
                                 2.34870489570875, 
                                 2.6140707088743675, 
                                 2.98, 
                                 3.0300000000000007, 
                                 3.595673308974897, 
                                 4.200249973963757, 
                                 4.79641814242165, 
                                 5.0, 
                                 5.385929291125633, 
                                 5.97])

    # wptherml includes zero-point energy but QuTip doesn't - we will subtract ZPE from ours
    test_1.energy_eigenvalues -= test_1.boson_energy_au / 2
    test_2.energy_eigenvalues -= test_2.boson_energy_au / 2
    test_3.energy_eigenvalues -= test_3.boson_energy_au / 2

    assert np.allclose(test_1.energy_eigenvalues, _expected_eigs_1)
    assert np.allclose(test_2.energy_eigenvalues, _expected_eigs_2)
    assert np.allclose(test_3.energy_eigenvalues, _expected_eigs_3)


def test_spin_boson_dynamics():
    test_args_1 = {
     "Number_of_Excitons": 1,
     "number_of_boson_levels": 2,
     "boson_energy_ev": 0.5 / 3.6749322175665e-2, 
     "exciton_energy_ev" : 0.5 / 3.6749322175665e-2, 
     "exciton_boson_coupling_ev" : 0.02 / 3.6749322175665e-2,
     "time_step_au" : 1.0
     
    }

    # test example vs QuTip J-C model for strong coupling between 1 spin and 2 level cavity
    # with spin = cavity omega = 0.5 atomic units, g = 0.02 (using RWA) and gamma = 5e-6 atomic units, kappa = 0
    # gamma is the spin decay rate in QuTip, kappa is the cavity decay rate
    # will set the initial condition corresponding to the spin in the excited state and cavity in the ground state

    # QuTip output for initial density matrix
    _expected_rho_init = np.array(
        [[0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
         [0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
         [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
         [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j]]
    )

    # QuTip output for density matrix after 1000 time steps with dt = 1 atomic unit
    _expected_rho_t_10 = np.array(
        [[4.93374249e-05+0.j,        0.00000000e+00+0.j, 0.00000000e+00+0.j,        0.00000000e+00+0.j       ],
         [0.00000000e+00+0.j,        9.60482179e-01+0.j,0.00000000e+00+0.1947019j, 0.00000000e+00+0.j       ],
         [0.00000000e+00+0.j,        0.00000000e+00-0.1947019j, 3.94684840e-02+0.j,        0.00000000e+00+0.j       ],
         [0.00000000e+00+0.j,        0.00000000e+00+0.j, 0.00000000e+00+0.j,        0.00000000e+00+0.j       ]]

    )

    sf = wptherml.SpectrumFactory()
    


    # instantiate 
    test_1 = sf.spectrum_factory("Spin-Boson", test_args_1)

    

    # set the rates in atomic units by hand - can set it in eV before instantiation, but too much trouble to convert...
    test_1.exciton_spontaneous_emission_rate_au = 5e-6

    # set initial condition for wptherml
    initial_cav = np.array([[1],[0]])
    initial_sp = np.array([[0],[1]])
    initial_ket = np.kron(initial_cav, initial_sp)

    # assign initial rho to test_1.rho attribute
    test_1.rho = initial_ket @ initial_ket.conj().T

    # test against QuTip initial rho for the same state
    assert np.allclose(test_1.rho, _expected_rho_init)

    # propagate for 1000 timesteps
    for i in range(10):
        test_1.rk4_update_on_rho()


    # test against QuTip rho at t = 10
    assert np.allclose(test_1.rho, _expected_rho_t_10)


# unit tests
def test_build_boson_basis():
    """
    Unit test for the build_boson_basis method.
    Tests against expected basis when the number of boson levels is 3 (basis includes |0>, |1>, |2>)
    """
    sf = wptherml.SpectrumFactory()
    args = {
        "number_of_boson_levels": 3,
    }

    test = sf.spectrum_factory("Spin-Boson", args)
    test.build_boson_basis()
    _expected_basis = np.eye(3) #np.matrix("1 0 0 ; 0 1 0 ; 0 0 1")
    assert np.allclose(_expected_basis, test.boson_basis)
    print(test.boson_basis)


def test_build_exciton_basis():
    """
    Unit test for the build_exciton_basis method.
    Tests against expected basis for three cases:

    (1) there is a single exciton
    (2) there are 2 excitons
    (3) there are 4 excitons
    """
    sf = wptherml.SpectrumFactory()

    # dictionaries for 3 cases
    args_1 = {
        "number_of_excitons": 1,
    }

    args_2 = {
        "number_of_excitons": 2,
    }

    args_3 = {
        "number_of_excitons": 4,
    }

    # instantiate 3 cases
    test_1 = sf.spectrum_factory("Spin-Boson", args_1)
    test_2 = sf.spectrum_factory("Spin-Boson", args_2)
    test_3 = sf.spectrum_factory("Spin-Boson", args_3)

    # build the exciton bases
    test_1.build_exciton_basis()
    test_2.build_exciton_basis()
    test_3.build_exciton_basis()

    _expected_one_exciton_basis = np.eye(2)
    _expected_two_exciton_basis = np.kron(
        _expected_one_exciton_basis, _expected_one_exciton_basis
    )
    _expected_four_exciton_basis = np.kron(
        _expected_one_exciton_basis,
        np.kron(_expected_one_exciton_basis, _expected_two_exciton_basis),
    )

    # for case 1 the single_exciton_basis should match the n_exciton_basis should match _expected_one_exciton_basis
    assert np.allclose(_expected_one_exciton_basis, test_1.single_exciton_basis)
    assert np.allclose(_expected_one_exciton_basis, test_1.n_exciton_basis)

    # for case 2, the single_exciton_basis should match the _expected_one_exciton_basis and the n_exciton_basis should match _expected_two_exciton_basis
    assert np.allclose(_expected_one_exciton_basis, test_2.single_exciton_basis)
    assert np.allclose(_expected_two_exciton_basis, test_2.n_exciton_basis)

    # for case 3, the single_exciton_basis should match the _expected_one_exciton_basis and the n_exciton_basis should match _expected_four_exciton_basis
    assert np.allclose(_expected_one_exciton_basis, test_3.single_exciton_basis)
    assert np.allclose(_expected_four_exciton_basis, test_3.n_exciton_basis)

    print(test_3.n_exciton_basis)
    print(_expected_four_exciton_basis)


def test_build_exciton_boson_basis():
    """
    Unit test for the build_exciton_boson_basis method.
    Tests against expected basis for three cases:

    (1) there is a single exciton and a boson with 2 levels
    (2) there is a single exciton and a boson with 3 levels
    (3) there are 2 excitons and a boson with 5 levels
    """
    sf = wptherml.SpectrumFactory()

    # dictionaries for 3 cases
    args_1 = {
        "number_of_excitons": 1,
        "number_of_boson_levels": 2,
    }

    args_2 = {
        "number_of_excitons": 1,
        "number_of_boson_levels": 3,
    }

    args_3 = {
        "number_of_excitons": 2,
        "number_of_boson_levels": 5,
    }

    # instantiate 3 cases
    test_1 = sf.spectrum_factory("Spin-Boson", args_1)
    test_2 = sf.spectrum_factory("Spin-Boson", args_2)
    test_3 = sf.spectrum_factory("Spin-Boson", args_3)

    # build the boson bases
    test_1.build_boson_basis()
    test_2.build_boson_basis()
    test_3.build_boson_basis()

    # build the exciton bases
    test_1.build_exciton_basis()
    test_2.build_exciton_basis()
    test_3.build_exciton_basis()

    # build the composite bases
    test_1.build_exciton_boson_basis()
    test_2.build_exciton_boson_basis()
    test_3.build_exciton_boson_basis()

    # create components of expected bases
    _expected_one_exciton_basis = np.eye(2)
    _expected_two_exciton_basis = np.kron(
        _expected_one_exciton_basis, _expected_one_exciton_basis
    )
    _expected_2_level_boson_basis = np.eye(2)
    _expected_3_level_boson_basis = np.eye(3)
    _expected_5_level_boson_basis = np.eye(5)

    # create expected composite basis for case 1
    _expected_exciton_boson_basis_case_1 = np.kron(
        _expected_2_level_boson_basis, _expected_one_exciton_basis
    )

    # create expected composite basis for case 2
    _expected_exciton_boson_basis_case_2 = np.kron(
        _expected_3_level_boson_basis, _expected_one_exciton_basis
    )

    # create expected composite basis for case 3
    _expected_exciton_boson_basis_case_3 = np.kron(
        _expected_5_level_boson_basis, _expected_two_exciton_basis
    )

    assert np.allclose(_expected_exciton_boson_basis_case_1, test_1.exciton_boson_basis)
    assert np.allclose(_expected_exciton_boson_basis_case_2, test_2.exciton_boson_basis)
    assert np.allclose(_expected_exciton_boson_basis_case_3, test_3.exciton_boson_basis)

    print(" Test 1 basis")
    print(test_1.exciton_boson_basis)
    print(" Expected test 1 basis")
    print(_expected_exciton_boson_basis_case_1)


def test_build_bosonic_ladder_operators():
    """
    Unit test for the build_bosonic_ladder_operator method

    (1) apply b to |1> when there is a 2-dimensional hilbert space and
        project onto <0|
    (2) apply b^+ to |0> when there is a 2-dimensional hilbert space and project
        onto <1|
    (3) apply b to |4> when there is a 6-dimensional hilbert space and
        project onto <3|
    (4) apply b^+ to |4> when there is a 6-dimensional hilbert space and
        project onto <5|
    (5) apply b^+ b to |8> when there is a 10 dimensional hilbert space and
        project onto <8|
    """
    # arguments for each case
    args_2d = {
        "number_of_boson_levels": 2,
    }
    args_6d = {
        "number_of_boson_levels": 6,
    }
    args_10d = {
        "number_of_boson_levels": 10,
    }

    sf = wptherml.SpectrumFactory()

    # instantiate cases
    test_2d = sf.spectrum_factory("Spin-Boson", args_2d)
    test_6d = sf.spectrum_factory("Spin-Boson", args_6d)
    test_10d = sf.spectrum_factory("Spin-Boson", args_10d)

    # build ladder oeprators
    test_2d.build_bosonic_ladder_operators()
    test_6d.build_bosonic_ladder_operators()
    test_10d.build_bosonic_ladder_operators()

    # create ket states for different cases
    _ket_1_2d = np.array([[0, 1]]).T #np.matrix("0 1").T
    _ket_0_2d = np.array([[1, 0]]).T #np.matrix("1 0").T

    _ket_4_6d = np.array([[0, 0, 0, 0, 1, 0]]).T #np.matrix("0 0 0 0 1 0").T

    _ket_8_10d = np.array([[0, 0, 0, 0, 0, 0, 0, 0, 1, 0]]).T #np.matrix("0 0 0 0 0 0 0 0 1 0").T

    _bra_0_2d = np.array([[1, 0]])
    _bra_1_2d = np.array([[0, 1]])

    _bra_3_6d = np.array([[0, 0, 0, 1, 0, 0]])   #np.matrix("0 0 0 1 0 0")
    _bra_5_6d = np.array([[0, 0, 0, 0, 0, 1]])   #np.matrix("0 0 0 0 0 1")

    _bra_8_10d = _ket_8_10d.T

    # compute <0|b|1> in 2D hilbert space
    _b_ket_1 = np.dot(test_2d.b_matrix, _ket_1_2d)
    _bra_0_2d_b_ket_1 = np.dot(_bra_0_2d, _b_ket_1)

    # compute <1|b^+|0> in 2D hilbert space
    _b_dag_ket_0 = np.dot(test_2d.b_dagger_matrix, _ket_0_2d)
    _bra_1_2d_b_ket_0 = np.dot(_bra_1_2d, _b_dag_ket_0)

    # both brackets should give 1
    assert np.isclose(_bra_0_2d_b_ket_1[0], 1.0)
    assert np.isclose(_bra_1_2d_b_ket_0[0], 1.0)

    # compute <3|b|4>
    _b_ket_4_6d = np.dot(test_6d.b_matrix, _ket_4_6d)
    _bra_3_6d_b_ket_4 = np.dot(_bra_3_6d, _b_ket_4_6d)

    # this bracket should give sqrt(4)
    assert np.isclose(_bra_3_6d_b_ket_4[0], 2.0)

    # compute <5|b^+|4>
    _b_dag_ket_4_6d = np.dot(test_6d.b_dagger_matrix, _ket_4_6d)
    _bra_5_6d_b_dag_ket_4 = np.dot(_bra_5_6d, _b_dag_ket_4_6d)

    # this bracket should give sqrt(5)
    assert np.isclose(_bra_5_6d_b_dag_ket_4[0], np.sqrt(5))

    # compute <8 | b^+ b | 8>
    b_dag_b_ket_8_10d = np.dot(
        test_10d.b_dagger_matrix, np.dot(test_10d.b_matrix, _ket_8_10d)
    )
    _bra_8_10d_b_dag_b_ket_8 = np.dot(_bra_8_10d, b_dag_b_ket_8_10d)

    # this bracket should be 8
    assert np.isclose(_bra_8_10d_b_dag_b_ket_8[0], 8.0)



def test_build_operator_for_exciton_j():
    """
        Unit test for the compute_operator_for_exciton_j method for the following test cases:

        (1a) There is 1 exciton system and we want to build the sigma^+ operator

        (1b) There is 1 exciton system and we want to build the sigma^- operator

        (1c) There is 1 exciton system and we want to build the sigma_z operator

        (2a) There are 2 exciton systems and we want to build sigma_z on the 1st one (index 0)

        (2b) There are 2 exciton systems and we want to build sigma_z on the 2nd one (index 1)

        (3a) There are 5 exciton systems and we want to build sigma_+ on the 3rd one (index 2)

        (3b) There are 5 exciton systems and we want to build sigma_- on the 2nd one (index 1)

    (3c) There are 5 exciton systems and we want to build sigma_x on the first one (index 0)

    (3d) There are 5 exciton systems and we want to build sigma_z on the last one (index 4)

    (3e) There are 5 exciton systems and we want to build sigma_+ sigma_- on the 4th one (index 3)

    """

    # dictionaries for case 1
    args_1 = {
        "number_of_excitons": 1,
    }
    # dictionaries for case 2
    args_2 = {
        "number_of_excitons": 2,
    }
    # dictionaries for case 3
    args_3 = {
        "number_of_excitons": 5,
    }

    # define some reference Pauli matrices
    _sigma_p = np.array([[0, 0], [1, 0]])   #np.matrix("0 0 ; 1 0")  # sigma_p |0> == sigma_p [1 0].T = |1> == [0 1].T
    _sigma_m = np.array([[0, 1], [0, 0]])   #np.matrix("0 1 ; 0 0")  # sigma_m |1> == sigma_m [0 1].T = |0> == [1 0].T
    _sigma_z = np.array([[1, 0], [0, -1]])  #np.matrix("1 0 ; 0 -1")
    _sigma_x = np.array([[0, 1], [1, 0]])   #np.matrix("0 1 ; 1 0")
    _ID = np.eye(2) # np.matrix("1 0 ; 0 1")
    _sigma_pm = np.array([[0, 0], [0, 1]])  #np.matrix("0 0 ; 0 1")

    sf = wptherml.SpectrumFactory()

    # instantiate case 1
    test_1 = sf.spectrum_factory("Spin-Boson", args_1)
    test_1.build_exciton_basis()

    # create sigma^+ operator and test
    test_1.build_operator_for_exciton_j(0, "sigma_p")
    assert np.allclose(test_1.exciton_operator_j, _sigma_p)

    # create sigma^- operator and test
    test_1.build_operator_for_exciton_j(0, "sigma_m")
    assert np.allclose(test_1.exciton_operator_j, _sigma_m)

    # create sigma_z operator and test
    test_1.build_operator_for_exciton_j(0, "sigma_z")
    assert np.allclose(test_1.exciton_operator_j, _sigma_z)

    # instantiate case 2
    test_2 = sf.spectrum_factory("Spin-Boson", args_2)
    test_2.build_exciton_basis()

    # create sigma^+ operator on 1st and test
    test_2.build_operator_for_exciton_j(0, "sigma_z")
    _expected_op_1 = np.kron(_sigma_z, _ID)
    assert np.allclose(test_2.exciton_operator_j, _expected_op_1)

    # create sigma^+ operator on 2nd and test
    test_2.build_operator_for_exciton_j(1, "sigma_z")
    _expected_op_2 = np.kron(_ID, _sigma_z)
    assert np.allclose(test_2.exciton_operator_j, _expected_op_2)

    # instantiate case 3
    test_3 = sf.spectrum_factory("Spin-Boson", args_3)
    test_3.build_exciton_basis()

    # create sigma^+ operator on 3rd and test
    test_3.build_operator_for_exciton_j(2, "sigma_p")
    _expected_op_1 = np.kron(_ID, np.kron(_ID, np.kron(np.kron(_sigma_p, _ID), _ID)))
    assert np.allclose(test_3.exciton_operator_j, _expected_op_1)

    # create sigma^m operator on 2nd and test
    test_3.build_operator_for_exciton_j(1, "sigma_m")
    _expected_op_2 = np.kron(_ID, np.kron(_sigma_m, np.kron(np.kron(_ID, _ID), _ID)))
    assert np.allclose(test_3.exciton_operator_j, _expected_op_2)

    # create sigma_x operator on first and test
    test_3.build_operator_for_exciton_j(0, "sigma_x")
    _expected_op_3 = np.kron(_sigma_x, np.kron(_ID, np.kron(np.kron(_ID, _ID), _ID)))
    assert np.allclose(test_3.exciton_operator_j, _expected_op_3)

    # create sigma_z operator on last and test
    test_3.build_operator_for_exciton_j(4, "sigma_z")
    _expected_op_4 = np.kron(_ID, np.kron(_ID, np.kron(np.kron(_ID, _ID), _sigma_z)))
    assert np.allclose(test_3.exciton_operator_j, _expected_op_4)

    # create sigma_+ sigma_- on the 4th one and test
    test_3.build_operator_for_exciton_j(3, "sigma_pm")
    _expected_op_5 = np.kron(_ID, np.kron(_ID, np.kron(np.kron(_ID, _sigma_pm), _ID)))
    assert np.allclose(test_3.exciton_operator_j, _expected_op_5)
    
def test_compute_exciton_boson_coupling_matrix():
    """
    Unit test for the compute_exciton_boson_coupling_matrix method for a 2-exciton 3-level bosonic system with \hbar \g = 0.01 eV.
    Compute matrix in the basis
        <q1, q2, s|H|q1',q2',s> =
    """

    # dictionaries for case 1
    args_1 = {
        "number_of_excitons": 1,
        "number_of_boson_levels": 3,
        "boson_energy_ev": 6.8028,
        "exciton_boson_coupling_ev" : 0.01
    }

    sf = wptherml.SpectrumFactory()

    # instantiate cases
    test_1 = sf.spectrum_factory("Spin-Boson", args_1)
    test_1.build_boson_basis()
    test_1.build_exciton_basis()
    test_1.build_exciton_boson_basis()
    test_1.build_exciton_boson_coupling_operator()

    _dim = 6
    _expected_matrix = np.zeros((_dim, _dim))

    # fill in boson energy elements using a loop over all basis states
    for i in range(3): # bra boson index
        for j in range(2): # bra exciton index
            _I = i * 2 + j
            for ip in range(3): # ket boson index
                for jp in range(2): # ket exciton index 
                    _J = ip * 2 + jp 
                    if i==(ip+1) and j==(jp-1):
                         _expected_matrix[_I, _J] = test_1.exciton_boson_coupling_au * np.sqrt(ip+1)
                    elif i==(ip-1) and j==(jp+1):
                         _expected_matrix[_I, _J] = test_1.exciton_boson_coupling_au * np.sqrt(ip)

    assert np.allclose(_expected_matrix, test_1.exciton_boson_coupling_operator)
    print("Expected Matrix")
    print(_expected_matrix)
    print("Built matrix")
    print(test_1.exciton_boson_coupling_operator)


test_build_boson_basis()
test_build_exciton_basis()
test_build_exciton_boson_basis()
test_build_bosonic_ladder_operators()
#test_compute_boson_energy_matrix()
test_build_operator_for_exciton_j()
test_compute_exciton_boson_coupling_matrix()



import numpy as np
from scipy.interpolate import UnivariateSpline


class Therml:
    """Collects methods for the computation of thermal radiative figures of merit

    Attributes
    ----------
    self.temperture : float
        temperature of the structure in Kelvin

    self.atmospheric_temperature : float
        temperature of the atmosphere in Kelvin

    self.lambda_bandgap : float
        bandgap wavelength of the target PV for PV or STPV in meters

    self.solar_angle : float
        angle of the incident solar spectrum in radians

    self.blackbody_spectrum : numpy array of floats
        Planck's blackbody spectrum for a given temperature

    self.thermal_emission_array : numpy array of floats
        thermal emission spectrum of structure for a given temperature

    self.blackbody_power_density : float
        total power density radiated into a hemisphere of a blackbody at a given temperature

    self.power_density : float
        total power density radiated by structure at a given temperature

    self.power_density_gradient : numpy array of floats (will be computed by this function)
        the gradient vector related to the total power density wrt changes in thicknesses of each layer

    self.photopic_luminosity_array : numpy array of floats
        the array of photopic luminosity values corresponding to each value of wavelength_array

    self.stpv_power_density : float (will be computed by this function)
        useful (sub-bandgap) power density radiated into a hemisphere

    self.stpv_power_density_gradient : numpy array of floats (will be computed by this function)
        the gradient vector related to the stpv power density wrt changes in thicknesses of each layer

    self.stpv_spectral_efficiency : float (will be computed by this function)
        spectral efficiency of an stpv emitter

    self.stpv_spectral_efficiency_gradient : numpy array of floats (will be computed by this function)
        the gradient vector related to the spectral efficiency wrt changes in thicknesses of each layer


    Returns
    -------
        None

    """

    def __init__(self, args):
        """constructor for the Therml class"""
        # parse args
        self._parse_therml_input(args)
        # self._compute_therml_spectrum()
        # self._compute_power_density()

    def _parse_therml_input(self, args):
        """method to parse the user inputs and define structures / simulation

        Returns
        -------
        None

        """
        if "temperature" in args:
            # user input expected in kelvin
            self.temperature = args["temperature"]

        else:
            # default temperature is 300 K
            self.temperature = 300

        if "bandgap wavelength" in args:
            self.lambda_bandgap = args["bandgap wavelength"]
        else:
            # default is ~InGaAsSb bandgap
            self.lambda_bandgap = 2254e-9

        if "atmospheric temperature" in args:
            self.atmospheric_temperature = args["atmospheric temperature"]
        else:
            self.atmospheric_temperature = 300

        if "solar angle" in args:
            self.solar_angle = args["solar angle"]
            self.solar_angle *= np.pi / 180
        else:
            self.solar_angle = 30 * np.pi / 180

    def _compute_therml_spectrum(self, wavelength_array, emissivity_array):
        """method to compute thermal emission spectrum of a structure

        Arguments
        ---------
        wavelength_array : numpy array of floats
            the array of wavelengths across which thermal emission spectrum will be computed

        emissivity_array : numpy array of floats
            the array of emissivity spectrum for the structure that you will compute the thermal emission of

        Attributes
        ----------
        blackbody_spectrum : numpy array of floats
            Planck's blackbody spectrum for a given temperature

        thermal_emission_array : numpy array of floats
            thermal emission spectrum of structure for a given temperature

        References
        ----------
        blackbody spectrum : Eq. (13) of https://github.com/FoleyLab/wptherml/blob/master/docs/Equations.pdf

        thermal_emission_array : Eq (12) of https://github.com/FoleyLab/wptherml/blob/master/docs/Equations.pdf
        with $\theta=0$
        """

        self.blackbody_spectrum = self._compute_blackbody_spectrum(
            wavelength_array, self.temperature
        )
        self.thermal_emission_array = self.blackbody_spectrum * emissivity_array

    def _compute_therml_spectrum_gradient(
        self, wavelength_array, emissivity_gradient_array
    ):
        """method to compute thermal emission spectrum of a structure

        Arguments
        ---------
        wavelength_array : numpy array of floats
            the array of wavelengths across which thermal emission spectrum will be computed

        emissivity_array : numpy array of floats
            the array of emissivity spectrum for the structure that you will compute the thermal emission of

        Attributes
        ----------
        blackbody_spectrum : numpy array of floats
            Planck's blackbody spectrum for a given temperature

        thermal_emission_array : numpy array of floats
            thermal emission spectrum of structure for a given temperature

        References
        ----------
        blackbody spectrum : Eq. (13) of https://github.com/FoleyLab/wptherml/blob/master/docs/Equations.pdf

        thermal_emission_array : Eq (12) of https://github.com/FoleyLab/wptherml/blob/master/docs/Equations.pdf
        with $\theta=0$
        """
        # get the dimension of the gradient vector
        _ngr = len(emissivity_gradient_array[0, :])
        # get the number of wavelengths
        _nwl = len(wavelength_array)
        # initialize the gradient array
        self.thermal_emission_gradient_array = np.zeros((_nwl, _ngr))

        self.blackbody_spectrum = self._compute_blackbody_spectrum(
            wavelength_array, self.temperature
        )

        for i in range(0, _ngr):
            self.thermal_emission_gradient_array[:, i] = (
                self.blackbody_spectrum * emissivity_gradient_array[:, i]
            )

    def _compute_blackbody_spectrum(self, wavelength_array, T):
        # speed of light in SI
        c = 299792458
        # plancks constant in SI
        h = 6.62607004e-34
        # boltzmanns constant in SI
        kb = 1.38064852e-23

        _bb_spectrum = 2 * h * c**2 / wavelength_array**5
        _bb_spectrum /= np.exp(h * c / (wavelength_array * kb * T)) - 1
        return _bb_spectrum

    def _compute_pv_stpv_power_density(self, wavelength_array):
        """method to compute the radiated power density of a PV-STPV structure specifically
        in the 0.3 - 0.5 eV range (~2450-4150 nm range)

        """
        # set lower- and upper limits on wavelength for integration
        _lambda_min = 2450e-9
        _lambda_max = 4150e-9

        # get the index associated with these upper- and lower-wavelengths
        _min_idx = np.abs(wavelength_array - _lambda_min).argmin()
        _max_idx = np.abs(wavelength_array - _lambda_max).argmin()

        # integrate the thermal emission spectrum over wavelength range using np.trapz
        self.pv_stpv_exciton_splitting_power = np.pi * np.trapz(
            self.thermal_emission_array[_min_idx:_max_idx],
            wavelength_array[_min_idx:_max_idx],
        )

    def _compute_power_density(self, wavelength_array):
        """method to compute the power density from blackbody spectrum and thermal emission spectrum

        Attributes
        ----------
        self.blackbody_power_density : float
            total power density radiated into a hemisphere of a blackbody at a given temperature

        self.power_density : float
            total power density radiated by structure at a given temperature

        References
        ----------
        Equation (15) and (16) of https://github.com/FoleyLab/wptherml/blob/master/docs/Equations.pdf

        """

        # integrate blackbody spectrum over wavelength using np.trapz
        self.blackbody_power_density = np.trapz(
            self.blackbody_spectrum, wavelength_array
        )

        # integrate the thermal emission spectrum over wavelength using np.trapz
        self.power_density = np.trapz(self.thermal_emission_array, wavelength_array)

        # account for angular integrals over hemisphere (assuming no angle dependence of emissivity)
        self.blackbody_power_density *= np.pi
        self.power_density *= np.pi

        # compute Blackbody power density from Stefan-Boltzmann law for validation
        # Stefan-Boltzmann constant
        sig = 5.670374419e-8
        self.stefan_boltzmann_law = sig * self.temperature**4

    def _compute_power_density_gradient(self, wavelength_array):
        """method to compute the gradient of the power density of a thermal emitter

        Arguments
        ---------
        wavelength_array : numpy array of floats
            the wavelengths over which the power density spectrum has been computed

        Attributes
        ----------

        self.power_density_gradient : numpy array of floats (will be computed by this function)
            the gradient vector related to the total power density wrt changes in thicknesses of each layer

        References
        ----------
        Equation (5) of https://journals.aps.org/prresearch/abstract/10.1103/PhysRevResearch.2.013018

        """
        _ngr = len(self.thermal_emission_gradient_array[0, :])
        # integrate the thermal emission spectrum over wavelength using np.trapz
        self.power_density_gradient = np.zeros(_ngr)

        for i in range(0, _ngr):
            self.power_density_gradient[i] = np.pi * np.trapz(
                self.thermal_emission_gradient_array[:, i], wavelength_array
            )

    def _compute_photopic_luminosity(self, wavelength_array):
        """computes the photopic luminosity function from a Gaussian fit

        Arguments
        ---------
        wavelength_array : numpy array of floats
            the array of wavelengths over which you will compute the photopic luminosity

        Attributes
        ----------
        self.photopic_luminosity_array : numpy array of floats
            the array of photopic luminosity values corresponding to each value of wavelength_array

        References
        ----------
        Data taken from http://www.cvrl.org/database/data/lum/linCIE2008v2e_5.htm
        """
        a = 1.02433
        b = 2.59462e14
        c = 5.60186e-07

        self._photopic_luminosity_array = a * np.exp(-b * (wavelength_array - c) ** 2)

    def _compute_stpv_power_density(self, wavelength_array):
        """method to compute the stpv power density from the thermal emission spectrum of a structure

        Arguments
        ---------
        wavelength_array : numpy array of floats
            the wavelengths over which the thermal emission spectrum is known

        Attributes
        ----------

        thermal_emission_array : numpy array of floats (already assigned)
            the thermal emission spectrum for each value of wavelength array for the stpv structure

        lambda_bandgap : float (already assigned)
            the bandgap wavelength -> upper limit on the integral for the stpv_power_density

        stpv_power_density : float (will be computed by this function)
            useful (sub-bandgap) power density radiated into a hemisphere


        References
        ----------
        Equation (17) of https://github.com/FoleyLab/wptherml/blob/master/docs/Equations.pdf

        """
        # compute the useful power density spectrum
        power_density_array = (
            self.thermal_emission_array * wavelength_array
        ) / self.lambda_bandgap

        # determine the index corresponding to lambda_bandgap in the wavelength_array
        # which will be used to determine the appropriate slice to feed to np.trapz
        bg_idx = np.abs(wavelength_array - self.lambda_bandgap).argmin()

        # integrate the power density between 0 to lambda_bandgap
        # by feeding the slice of the power_density_array and wavelength_array
        # from 0:bg_idx to the trapz function
        self.stpv_power_density = np.pi * np.trapz(
            power_density_array[:bg_idx], wavelength_array[:bg_idx]
        )

    def _compute_stpv_power_density_gradient(self, wavelength_array):
        """method to compute the power density from blackbody spectrum and thermal emission spectrum
        Arguments
        ---------
        wavelength_array : numpy array of floats
            the wavelengths over which the power density spectrum has been computed

        Attributes
        ----------
        self.stpv_power_density_gradient : numpy array of floats (will be computed by this function)
            the gradient vector related to the stpv power density wrt changes in thicknesses of each layer

        References
        ----------
        Equation (5) of https://journals.aps.org/prresearch/abstract/10.1103/PhysRevResearch.2.013018

        """
        _ngr = len(self.thermal_emission_gradient_array[0, :])
        # integrate the thermal emission spectrum over wavelength using np.trapz
        self.stpv_power_density_gradient = np.zeros(_ngr)
        # compute the useful power density spectrum

        for i in range(0, _ngr):
            stpv_power_density_array_prime = (
                self.thermal_emission_gradient_array[:, i]
                * wavelength_array
                / self.lambda_bandgap
            )
            self.stpv_power_density_gradient[i] = np.pi * np.trapz(
                stpv_power_density_array_prime, wavelength_array
            )

    def _compute_stpv_spectral_efficiency(self, wavelength_array):
        """method to compute the stpv spectral efficiency from the thermal emission spectrum of a structure

        Arguments
        ---------
        wavelength_array : numpy array of floats
            the wavelengths over which the thermal emission spectrum is known

        Attributes
        ----------

        thermal_emission_array : numpy array of floats (already assigned)
            the thermal emission spectrum for each value of wavelength array for the stpv structure

        lambda_bandgap : float (already assigned)
            the bandgap wavelength -> upper limit on the integral for the stpv_power_density

        stpv_power_density : float (will be computed by this function)
            useful (sub-bandgap) power density radiated into a hemisphere

        power_density : float (will be computed by this function)
            total power density radiated into a hemisphere

        stpv_spectral_efficiency : float (will be computed by this function)
            spectral efficiency of an stpv emitter

        References
        ----------
        Equation (18) of https://github.com/FoleyLab/wptherml/blob/master/docs/Equations.pdf

        """

        self._compute_stpv_power_density(wavelength_array)
        self._compute_power_density(wavelength_array)
        self.stpv_spectral_efficiency = self.stpv_power_density / self.power_density

    def _compute_stpv_spectral_efficiency_gradient(self, wavelength_array):
        """method to compute the gradient of the
           stpv spectral efficiency from the thermal emission spectrum of a structure

        Arguments
        ---------
        wavelength_array : numpy array of floats
            the wavelengths over which the thermal emission spectrum is known

        Attributes
        ----------
        self.stpv_spectral_efficiency_gradient : numpy array of floats (will be computed by this function)
            the gradient vector related to the spectral efficiency wrt changes in thicknesses of each layer

        References
        ----------
        Equation (4) of https://journals.aps.org/prresearch/abstract/10.1103/PhysRevResearch.2.013018

        """
        # get the number of elements in the gradient
        _ngr = len(self.thermal_emission_gradient_array[0, :])

        # initialize the gradient array
        self.stpv_spectral_efficiency_gradient = np.zeros(_ngr)

        # using the notation from Eq. (4)
        # from https://journals.aps.org/prresearch/abstract/10.1103/PhysRevResearch.2.013018
        self._compute_stpv_power_density(wavelength_array)
        self._compute_power_density(wavelength_array)

        _P = self.power_density
        _rho = self.stpv_power_density

        # determine the index corresponding to lambda_bandgap in the wavelength_array
        # which will be used to determine the appropriate slice to feed to np.trapz
        _bg_idx = np.abs(wavelength_array - self.lambda_bandgap).argmin()

        for i in range(0, _ngr):
            _rho_prime_integrand = (
                self.thermal_emission_gradient_array[:_bg_idx, i]
                * wavelength_array[:_bg_idx]
                / self.lambda_bandgap
            )
            _rho_prime = np.pi * np.trapz(
                _rho_prime_integrand, wavelength_array[:_bg_idx]
            )
            _P_prime = np.pi * np.trapz(
                self.thermal_emission_gradient_array[:, i], wavelength_array
            )
            self.stpv_spectral_efficiency_gradient[i] = (
                _rho_prime * _P - _P_prime * _rho
            ) / (_P * _P)

    def _compute_pv_short_circuit_current(
        self, wavelength_array, absorptivity_array, spectral_response, solar_spectrum
    ):
        """method to approximate the short circuit current of a PV cell

        Arguments
        ---------
        wavelength_array : numpy array of floats
            the wavelengths over which the absorptivity of the structure is known

        absorptivity_array : numpy array of floats
            the absorptivity / emissivity spectrum of the PV structure

        spectral_response : numpy array of floats
            the mapping between power in  (W) / current out (Amps) of the active layer

        solar_spectrum : numpy array of floats
            the solar spectrum incident on the PV structure

        Attributes
        ----------
        self.short_circuit_current : float
            the short circuit current in Amps / m^2 of the surface

        Returns
        -------
        None

        """
        self.pv_short_circuit_current = np.trapz(
            absorptivity_array * spectral_response * solar_spectrum, wavelength_array
        )

    def _compute_luminous_efficiency(self, wavelength_array):
        """method to compute the luminous efficiency for an incandescent from the thermal emission spectrum of a structure

        Arguments
        ---------
        wavelength_array : numpy array of floats
            the wavelengths over which the thermal emission spectrum is known

        Attributes
        ----------

        self.thermal_emission_array : numpy array of floats (already assigned)
            the thermal emission spectrum for each value of wavelength array for the stpv structure

        self.photopic_luminosity_array : numpy array of floats (can be computed by calling self._compute_photopic_luminosity(wavelength_array))
            photopic luminosity function values corresponding to wavelength_array

        self.luminous_efficiency : float (will be computed by this function)
            the spectral efficiency of the incandescent source

        References
        ----------
        Equation (27) of https://github.com/FoleyLab/wptherml/blob/master/docs/Equations.pdf

        """
        # self._compute_therml_spectrum(wavelength_array, emissivity_array)
        self._compute_photopic_luminosity(wavelength_array)
        vl = self._photopic_luminosity_array
        TE = self.thermal_emission_array

        Numerator = np.trapz(vl * TE, wavelength_array)
        Denominator = np.trapz(TE, wavelength_array)

        self.luminous_efficiency = Numerator / Denominator

    def _compute_thermal_radiated_power(
        self,
        emissivity_array_s,
        emissivity_array_p,
        theta_vals,
        theta_weights,
        wavelength_array,
    ):
        """Method to compute the power radiated by the cooling structure

        Attributes
        ----------
        None

        Returns
        -------
        P_rad

        References
        ----------
        See Eq. (2) of https://www.nature.com/articles/nature13883
        """

        num_angles = len(theta_vals)
        self._compute_therml_spectrum(wavelength_array, emissivity_array_s[0, :])

        # loop over angles
        P_rad = 0.0
        for i in range(0, num_angles):
            _TE = (
                self.blackbody_spectrum
                * np.cos(theta_vals[i])
                * 0.5
                * (emissivity_array_p[i, :] + emissivity_array_s[i, :])
            )
            _TE_INT = np.trapz(_TE, wavelength_array)
            P_rad += _TE_INT * np.sin(theta_vals[i]) * theta_weights[i]

        P_rad *= np.pi * 2

        return P_rad

    def _compute_thermal_radiated_power_gradient(
        self,
        emissivity_gradient_array_s,
        emissivity_gradient_array_p,
        theta_vals,
        theta_weights,
        wavelength_array,
    ):
        """Method to compute the gradient of the power radiated by the cooling structure

        Attributes
        ----------
        None

        Returns
        -------
        _emitted_thermal_spectrum_gradient

        References
        ----------
        See Eq. (2) of https://www.nature.com/articles/nature13883

        """
        num_angles = len(theta_vals)
        # we don't care about the emissivity - just calling this for the blackbody spectrum
        self._compute_therml_spectrum(
            wavelength_array, emissivity_gradient_array_p[0, :, 0]
        )

        _ngr = len(emissivity_gradient_array_s[0, 0, :])
        _nth = len(emissivity_gradient_array_s[:, 0, 0])

        # instantiate P_rad_prime
        _emitted_thermal_spectrum_gradient = np.zeros(_ngr)
        for i in range(0, _ngr):
            _P_rad_prime = 0
            for j in range(0, _nth):
                _TE = (
                    self.blackbody_spectrum
                    * np.cos(theta_vals[j])
                    * 0.5
                    * (
                        emissivity_gradient_array_p[j, :, i]
                        + emissivity_gradient_array_s[j, :, i]
                    )
                )
                _TE_INT = np.trapz(_TE, wavelength_array)

                _P_rad_prime += (
                    2 * np.pi * _TE_INT * np.sin(theta_vals[j]) * theta_weights[j]
                )
            _emitted_thermal_spectrum_gradient[i] = _P_rad_prime

        return _emitted_thermal_spectrum_gradient

    def _compute_atmospheric_radiated_power(
        self,
        atmospheric_transmissivity,
        emissivity_array_s,
        emissivity_array_p,
        theta_vals,
        theta_weights,
        wavelength_array,
    ):
        """Method to compute the power radiated by the atmosphere and absorbed by cooling structure

        Attributes
        ----------
        None

        Returns
        -------
        P_atm

        References
        ----------
        See Eq. (3) of https://www.nature.com/articles/nature13883

        """
        num_angles = len(theta_vals)

        # make sure we are getting the blackbody spectrum of the atmosphere
        # store the structure temperature
        _T_temp = self.temperature
        # update the structure temperature, which is the attribute that
        # is used for computing the blackbody spectrum in _compute_therml_spectrum
        self.temperature = self.atmospheric_temperature
        self._compute_therml_spectrum(wavelength_array, emissivity_array_s[0, :])
        # set the structure temperature back to _T_temp in case
        # one wants to compute the thermal emission of the structure again!
        self.temperature = _T_temp
        P_atm = 0.0
        for i in range(0, num_angles):
            # get the term that goes in the exponent of the atmospheric transmissivity
            _o_over_cos_t = 1 / np.cos(theta_vals[i])
            _emissivity_atm = (
                np.ones(len(atmospheric_transmissivity))
                - atmospheric_transmissivity**_o_over_cos_t
            )
            _TE_atm = self.blackbody_spectrum * _emissivity_atm * np.cos(theta_vals[i])
            _absorbed_TE_spectrum = (
                _TE_atm * 0.5 * (emissivity_array_p[i, :] + emissivity_array_s[i, :])
            )
            _absorbed_TE = np.trapz(_absorbed_TE_spectrum, wavelength_array)
            P_atm += _absorbed_TE * np.sin(theta_vals[i]) * theta_weights[i]
        P_atm *= 2 * np.pi

        return P_atm

    def _compute_atmospheric_radiated_power_gradient(
        self,
        atmospheric_transmissivity,
        emissivity_gradient_array_s,
        emissivity_gradient_array_p,
        theta_vals,
        theta_weights,
        wavelength_array,
    ):
        """Method to compute the gradient of the power radiated by the atmosphere and absorbed by the cooling structure

        Attribute
        ---------
        None

        Returns
        -------
        _absorbed_solar_spectrum_gradient
        """

        _nth = len(theta_vals)
        _ngr = len(emissivity_gradient_array_p[0, 0, :])
        _absorbed_atmospheric_radiation_gradient = np.zeros(_ngr)

        # make sure we are getting the blackbody spectrum of the atmosphere
        # store the structure temperature
        _T_temp = self.temperature
        # update the structure temperature, which is the attribute that
        # is used for computing the blackbody spectrum in _compute_therml_spectrum
        self.temperature = self.atmospheric_temperature
        self._compute_therml_spectrum(
            wavelength_array, emissivity_gradient_array_p[0, :, 0]
        )
        # set the structure temperature back to _T_temp in case
        # one wants to compute the thermal emission of the structure again!
        self.temperature = _T_temp

        for i in range(0, _ngr):
            P_atm_prime = 0
            for j in range(0, _nth):
                # get the term that goes in the exponent of the atmospheric transmissivity
                _o_over_cos_t = 1 / np.cos(theta_vals[j])
                _emissivity_atm = (
                    np.ones(len(atmospheric_transmissivity))
                    - atmospheric_transmissivity**_o_over_cos_t
                )
                _TE_atm = (
                    self.blackbody_spectrum * _emissivity_atm * np.cos(theta_vals[j])
                )
                _absorbed_TE_spectrum = (
                    _TE_atm
                    * 0.5
                    * (
                        emissivity_gradient_array_p[j, :, i]
                        + emissivity_gradient_array_s[j, :, i]
                    )
                )
                _absorbed_TE = np.trapz(_absorbed_TE_spectrum, wavelength_array)
                P_atm_prime += (
                    2 * np.pi * _absorbed_TE * np.sin(theta_vals[j]) * theta_weights[j]
                )
            _absorbed_atmospheric_radiation_gradient[i] = P_atm_prime

        return _absorbed_atmospheric_radiation_gradient

    def _compute_solar_radiated_power(
        self, solar_spectrum, emissivity_array_s, emissivity_array_p, wavelength_array
    ):
        """Method to compute the solar power absorbed by the cooling structure

        Attributes
        ----------
        None

        Returns
        -------
        P_sun

        References
        ----------
        See Eq. (4) of https://www.nature.com/articles/nature13883
        """
        # compute the absorbed solar spectrum
        _absorbed_solar_spectrum = (
            solar_spectrum * 0.5 * (emissivity_array_p + emissivity_array_s)
        )
        # integrate it!
        P_sun = np.trapz(_absorbed_solar_spectrum, wavelength_array)
        return P_sun

    def _compute_solar_radiated_power_gradient(
        self,
        solar_spectrum,
        emissivity_gradient_array_s,
        emissivity_gradient_array_p,
        wavelength_array,
    ):
        """Method to compute the gradient of the solar power absorbed by the cooling structure

        Attributes
        ----------
        None

        Returns
        -------
        _absorbed_solar_spectrum_gradient

        References
        ----------
        See Eq. (4) of https://www.nature.com/articles/nature13883
        """
        # get the dimension of the gradient vector
        _ngr = len(emissivity_gradient_array_s[0, :])
        _absorbed_solar_spectrum_gradient = np.zeros(_ngr)
        for i in range(0, _ngr):
            _absorbed_solar_spectrum_prime = (
                solar_spectrum
                * 0.5
                * (
                    emissivity_gradient_array_p[:, i]
                    + emissivity_gradient_array_s[:, i]
                )
            )
            # integrate it!
            _absorbed_solar_spectrum_gradient[i] = np.trapz(
                _absorbed_solar_spectrum_prime, wavelength_array
            )
        return _absorbed_solar_spectrum_gradient



"""
wptherml
A python package for modeling light-matter interactions!
"""

# Add imports here
from .spectrum_driver import SpectrumDriver
from .em import TmmDriver
from .mie import MieDriver
from .therml import Therml
from .exciton import ExcitonDriver
from .factory import SpectrumFactory
from .materials import Materials
from .optdriver import OptDriver
from .spin_boson import SpinBosonDriver

# Handle versioneer
from ._version import get_versions

versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions



import numpy as np
from scipy.optimize import minimize
from scipy.optimize import basinhopping
import time
from .em import TmmDriver
#import torch
#import torch.nn as nn
#import torch.optim as optim
#import tqdm
#from sklearn.model_selection import train_test_split
#from sklearn.preprocessing import StandardScaler
#from sklearn.preprocessing import MinMaxScaler
import copy
#from pyqubo import Binary
#import neal
#import pandas as pd


class OptDriver(TmmDriver):
    """Compute the absorption, scattering, and extinction spectra of a sphere using Mie theory

    Attributes
    ----------
    TBD

    Returns
    -------
    None

    Examples
    --------
    >>> fill_in_with_actual_example!
    """

    def __init__(self, args):
        # store args
        self.args = args

        self.start_time = time.time()

        # ordinary tmm driver input
        args = {k.lower(): v for k, v in args.items()}
        self.parse_input(args)
        self.parse_optimization_input(args)
        self.set_refractive_index_array()
        # compute reflectivity spectrum
        self.compute_spectrum()
        # print("We started optimizing")
        # self.optimize()

    def parse_optimization_input(self, args):
        """Parse additional options related to the optimization, including:
        - the objective function to be optimized
        - minimization or maximization of the objective
        - the optimization method to be used
        - lower- and upper-bounds on the layer thicknesses

        """
        # which objective function do we want to optimize
        if "objective_function" in args:
            self.objective_function = args["objective_function"]
        # default to selective mirror
        else:
            self.objective_function = "selective_mirror"

        # do we want to minimze the objective?
        if "minimization" in args:
            self.mimimization = args["minimization"]
        # usually we want to mazimize, so default is false
        else:
            self.minimization = False

        # set bounds on thickness of each layer
        if "lower_bound" in args:
            self.lower_bound = args["lower_bound"]
        else:
            self.lower_bound = 1

        if "upper_bound" in args:
            self.upper_bound = args["upper_bound"]
        else:
            self.upper_bound = 1000

        if "combinatorial_optimization" in args:
            self.combinatorial_optimization = True
            self.qubo_thickness_optimization = False
            self.optimization_dictionary = args["optimization_dictionary"]
        if "qubo_thickness_optimization" in args:
            self.qubo_thickness_optimization = True
            self.combinatorial_optimization = False
            self.optimization_dictionary = args["optimization_dictionary"]

        if "random_perturbation_scale" in args:
            self.random_perturbation_scale = args["random_perturbation_scale"]

        else:
            # default will be 10%, so we can randomly perturb a given layer by +/- 10% 
            # between basin hopping cycles
            self.random_perturbation_scale = 0.1

    def optimize_qubo(self, fom_function):
        if self.combinatorial_optimization == True:
            self.optimizer = self._qubo_combinatorial_structure_optimization(
                self, self.optimization_dictionary, fom_function
            )
            self.optimizer.learning_loop(num_to_train=25, num_iterations=10000)

        if self.qubo_thickness_optimization == True:
            self.optimizer = self._qubo_thickness_and_alloy_optimization(
                self,
                optimization_dict=self.optimization_dictionary,
                fom_func=fom_function,
            )
            self.optimizer.learning_loop(
                num_to_train=100,
                n_epochs=250,
                l2_lambda=0.001,
                num_iterations=400,
                reduction_factor=2,
                l1_lambda=0.000001,
                K=25,
                LR=0.01,
            )

    def optimize_bfgs(self):
        """ "
        Method to wrap the l-bfgs-b optimizer
        """
        # initialize x array
        x_start = self.thickness_array[1 : self.number_of_layers - 1] * 1e9

        # set bounds
        bfgs_xmin = self.lower_bound * np.ones(self.number_of_layers - 2)
        bfgs_xmax = self.upper_bound * np.ones(self.number_of_layers - 2)
        # rewrite the bounds in the way required by L-BFGS-B
        bfgs_bounds = [(low, high) for low, high in zip(bfgs_xmin, bfgs_xmax)]

        fom_start, grad_start = self.compute_fom_and_gradient_from_thickness_array(
            x_start
        )
        print(f" Initial FOM is {fom_start}")
        print(f" Initial Gradient is {grad_start}")

        ret = minimize(
            self.compute_fom_and_gradient_from_thickness_array,
            x_start,
            method="L-BFGS-B",
            jac=True,
            bounds=bfgs_bounds,
        )
        print(ret.x)

    def optimize_basin_hopping(self):
        """
        Method to wrap the l-bfgs-b optimizer
        """
        x_start = self.thickness_array[1 : self.number_of_layers - 1] * 1e9

        # set bounds for L-BFGS-B local optimizations
        bfgs_xmin = self.lower_bound * np.ones(self.number_of_layers - 2)
        bfgs_xmax = self.upper_bound * np.ones(self.number_of_layers - 2)

        # rewrite the bounds in the way required by L-BFGS-B
        bfgs_bounds = [(low, high) for low, high in zip(bfgs_xmin, bfgs_xmax)]

        fom_start, grad_start = self.compute_fom_and_gradient_from_thickness_array(
            x_start
        )

        # should update to determine if "jac" : True is consistent with 
        # selected FOM
        minimizer_kwargs = {"method": "L-BFGS-B", "jac": True, "bounds": bfgs_bounds}

        print(f" Initial FOM is {fom_start}")
        print(f" Initial Gradient is {grad_start}")
        ret = basinhopping(
            func = self.compute_fom_and_gradient_from_thickness_array,
            x0 = x_start,
            minimizer_kwargs=minimizer_kwargs,
            niter=50,
            take_step=self._take_step,
            callback=self._print_callback,
        )
        return ret

    def _take_step(self,x):
        """ method to apply random perturbations to 
            the current structure x in the basin hopping routine
        """

        _x_perturbed = np.copy(x)

        dim = len(_x_perturbed)
        for i in range(0,dim):
            _x_curr = _x_perturbed[i]
            if _x_curr * self.random_perturbation_scale > 1.0:
                _x_range = int(_x_curr * self.random_perturbation_scale)
            else:
                _x_range = 1
            _pert = np.random.randint(-_x_range, _x_range)
            # we don't want to perturb to less than the lower bound or greater than the upper bound
            _x_try = _x_curr + _pert 
            if _x_try > self.lower_bound and _x_try < self.upper_bound:
                _x_perturbed[i] = _x_try
            elif _x_try <= self.lower_bound: 
                _x_perturbed[i] = _x_curr + np.abs(_pert) #<== increase in thickness
            elif _x_try >= self.upper_bound:
                _x_perturbed[i] = _x_curr - np.abs(_pert) #<== decrease in thickness
            else:
                _x_perturbed[i] = _x_curr #<== do not change
            
        return _x_perturbed
    
    def _print_callback(self,x, f, accepted):
        c_time = time.time()
        print(f"\n Time elapsed is {c_time-self.start_time}")
        print(f" Current structure is {x}")
        print(f" Current FOM is {f}")

    def compute_fom_and_gradient_from_thickness_array(self, x0):
        """ "
        Method to update the thickness array, the relevant spectra, objective function, and (if supported)
        the gradient
        """
        print(" Computing figure of merit and graident from thickness array ", x0)

        # need to add options to use other figures of merit
        self.thickness_array[1 : self.number_of_layers - 1] = x0 * 1e-9
        self.compute_spectrum()
        self.compute_selective_mirror_fom()
        # compute_selective_mirror_fom_gradient calls compute_spectrum_gradient()
        self.compute_selective_mirror_fom_gradient()

        fom_1 = self.reflection_efficiency
        grad_1 = self.reflection_efficiency_gradient

        fom_2 = self.reflection_selectivity
        grad_2 = self.reflection_selectivity_gradient

        fom_3 = self.transmission_efficiency
        grad_3 = self.transmission_efficiency_gradient

        _expected_fom = self.reflection_efficiency_weight * fom_1 
        _expected_fom += self.reflection_selectivity_weight * fom_2
        _expected_fom += self.transmission_efficiency_weight * fom_3 

        fom = self.selective_mirror_fom
        grad = self.selective_mirror_fom_gradient

        assert np.isclose(_expected_fom, fom)

        if self.minimization:
            fom *= 1
            grad *= 1e-9
        else:
            fom *= -1
            grad *= -1e-9

        print(f" Current FOM is {fom}")
        print(f" Current Gradient is {grad}")
        return fom, grad



"""
Unit and regression test for the wpspec package.
"""

# Import package, test suite, and other packages as needed
import wptherml
import numpy as np
import pytest
import sys


""" perform a series of tests that would apply to the simple
    case of a r = 100e-9 m spehere made of glass with ri = 1.5+0j
    in air with ri = 1.0+0j
"""
sf = wptherml.SpectrumFactory()
testargs = {
    "radius": 100e-9,
    "wavelength_list": [500e-9, 500e-9, 1],
    "sphere_material": "sio2",
    "medium_material": "air",
}
# create test instance of mie
mietest = sf.spectrum_factory("Mie", testargs)
# hard-code relative refractive index to be 1.5+0j
mietest._relative_refractive_index_array[:] = 1.5 + 0j
# create test instance of coefficient array
mietest._compute_n_array(mietest._size_factor_array[0])


def test_compute_mie_coefficients():
    # sf = wptherml.SpectrumFactory()
    # mt = sf.spectrum_factory('Mie', r)
    # mt._compute_mie_coeffients(ri, mu, x)

    expected_an = np.array(
        [
            1.09393760e-01 - 3.12132609e-01j,
            9.58682444e-04 - 3.09477523e-02j,
            1.63054752e-06 - 1.27692790e-03j,
            9.66113753e-10 - 3.10823704e-05j,
            2.39854703e-13 - 4.89749633e-07j,
            2.88856083e-17 - 5.37453331e-09j,
            1.88087968e-21 - 4.33691097e-11j,
        ]
    )

    expected_bn = np.array(
        [
            8.51059716e-03 - 9.18594954e-02j,
            1.35466949e-05 - 3.68055857e-03j,
            8.15604142e-09 - 9.03108042e-05j,
            2.07821108e-12 - 1.44160018e-06j,
            2.56048144e-16 - 1.60015044e-08j,
            1.69880890e-20 - 1.30338363e-10j,
            6.58730141e-25 - 8.11621920e-13j,
        ]
    )

    expected_cn = np.array(
        [
            1.02133636 + 9.46247560e-02j,
            0.5558344 + 2.04580879e-03j,
            0.34394533 + 3.10619798e-05j,
            0.22121405 + 3.18902215e-07j,
            0.14432436 + 2.30940695e-09j,
            0.09482464 + 1.23592890e-11j,
            0.0625545 + 5.07706007e-14j,
        ]
    )
    expected_dn = np.array(
        [
            0.94943955 + 3.32752039e-01j,
            0.57237172 + 1.77306162e-02j,
            0.33722182 + 4.30608658e-04j,
            0.21300494 + 6.62069833e-06j,
            0.13775492 + 6.74654196e-08j,
            0.09001634 + 4.83795816e-10j,
            0.05915672 + 2.56557430e-12j,
        ]
    )
    check_lst = [expected_an, expected_bn, expected_cn, expected_dn]
    # for i in check_lst:
    m_val = mietest._relative_refractive_index_array[0]
    mu_val = mietest._relative_permeability
    x_val = mietest._size_factor_array[0]

    mietest._compute_mie_coeffients(m_val, mu_val, x_val)
    assert np.allclose(mietest._an, expected_an, 1e-5)
    assert np.allclose(mietest._bn, expected_bn, 1e-5)
    assert np.allclose(mietest._cn, expected_cn, 1e-5)
    assert np.allclose(mietest._dn, expected_dn, 1e-5)


def test_compute_s_jn():
    """tests private method in MieDriver _compute_s_jn(n, z)"""

    expected_result = np.array(
        [
            3.56355664e-01,
            9.39097524e-02,
            1.72993686e-02,
            2.45504861e-03,
            2.83621841e-04,
            2.76413774e-05,
            2.33017989e-06,
        ]
    )

    result = mietest._compute_s_jn(mietest._n_array, mietest._size_factor_array[0])

    assert np.allclose(result, expected_result, 1e-5)


def test_compute_s_yn():
    """test private method in MieDriver _compute_s_yn(n, z)"""

    expected_result = np.array(
        [
            -9.52514026e-01,
            -2.02805182e00,
            -7.11684779e00,
            -3.76158009e01,
            -2.62286481e02,
            -2.25831465e03,
            -2.31001396e04,
        ]
    )

    result = mietest._compute_s_yn(mietest._n_array, mietest._size_factor_array[0])

    assert np.allclose(result, expected_result, 1e-5)


def test_compute_s_hn():
    """test private method in MieDriver _compute_s_hn(n, z)"""

    expected_result = np.array(
        [
            3.56355664e-01 - 9.52514026e-01j,
            9.39097524e-02 - 2.02805182e00j,
            1.72993686e-02 - 7.11684779e00j,
            2.45504861e-03 - 3.76158009e01j,
            2.83621841e-04 - 2.62286481e02j,
            2.76413774e-05 - 2.25831465e03j,
            2.33017989e-06 - 2.31001396e04j,
        ]
    )

    result = mietest._compute_s_hn(mietest._n_array, mietest._size_factor_array[0])

    assert np.allclose(result, expected_result, 1e-5)


def test_compute_z_jn_prime():
    """test private method in MieDriver _compute_z_jn_prime(n,z)"""

    expected_result = np.array(
        [
            5.94700852e-01,
            2.59990229e-01,
            6.61123694e-02,
            1.19188333e-02,
            1.66699586e-03,
            1.90561452e-04,
            1.84239201e-05,
        ]
    )

    result = mietest._compute_z_jn_prime(
        mietest._n_array, mietest._size_factor_array[0]
    )


def test_compute_z_hn_prime():
    """test private method in MieDriver _compute_z_hn_prime(n, z)"""

    expected_result = np.array(
        [
            5.94700852e-01 + 6.43497032e-01j,
            2.59990229e-01 + 2.85913922e00j,
            6.61123694e-02 + 1.88020183e01j,
            1.19188333e-02 + 1.41519909e02j,
            1.66699586e-03 + 1.26416300e03j,
            1.90561452e-04 + 1.32202890e04j,
            1.84239201e-05 + 1.58863095e05j,
        ]
    )

    result = mietest._compute_z_hn_prime(
        mietest._n_array, mietest._size_factor_array[0]
    )

    assert np.allclose(result, expected_result, 1e-5)


def test_compute_q_scattering():
    """test private method in MieDriver _compute_q_scattering(n, z)"""

    expected_result = 0.4541540910257134

    result = mietest._compute_q_scattering(mietest._size_factor_array[0])
    assert np.isclose(result, expected_result, 1e-5)


def test_compute_q_extinction():
    """test private method in MieDriver _compute_q_extinction(n, z)"""

    expected_result = 0.4541540910257134

    result = mietest._compute_q_extinction(mietest._size_factor_array[0])
    assert np.isclose(result, expected_result, 1e-5)



import numpy as np
from matplotlib import pyplot as plt
from .spectrum_driver import SpectrumDriver


class ExcitonDriver(SpectrumDriver):
    """A class for computing the dynamics and spectra of a system modelled by the Frenkel Exciton Hamiltonian

    Attributes
    ----------
    radius : float
        the radius of the sphere
    
    c_vector : numpy array of flaots
        vector coefficient associated with the weight of a state in an expansion of the wavefunction

    density_matrix : numpy array of floats
        Adjoint of a c_vector multiplied by the original c_vector

    exciton_energy : float
        energy of the monomer exciton in atomic units

        aggregate_shape : tuple
            The number of monomers along each coordinate (Nx, Ny, Nz)

    displacement_vector : numpy array of flaots
        displacement between each monomer in cartesian coordinates

    transition_dipole_moment : numpy array of floats
        dipole associated with the transition from the ground state to the excited state in atomic units

    refractice_index : float
        refractive index of the monomer film


    

    Returns
    -------
    None

    Examples
    --------
    >>> fill_in_with_actual_example!
    """

    def __init__(self, args):
        self.parse_input(args)
        # print("Exciton Energy is  ", self.exciton_energy)
        # allocate the exciton Hamiltonian

        self.exciton_hamiltonian = np.zeros(
            (self.number_of_monomers, self.number_of_monomers)
        )
        self.c_vector = np.zeros(
            (self.number_of_monomers, 1),  dtype=complex
        )  # <== wavefunction coefficient vector
        
        # define cartesian coordinates of each monomer
        self._compute_cartesian_coordinates()

        # build exciton Hamiltonian
        self.build_exciton_hamiltonian()

        # Probably want to allow the user to specify an initial state!
        # but right now just have the initial state with exciton localized on site 1
        self.c_vector[0,0] = 1 + 0j
        self.density_matrix = np.dot(self.c_vector, np.conj(self.c_vector.T))

    def parse_input(self, args):
        if "exciton_energy" in args:
            self.exciton_energy = args["exciton_energy"]
        else:
            self.exciton_energy = 0.5

        if "aggregate_shape" in args:
            self.aggregate_shape = args["aggregate_shape"]

        # default shape is (2, 2, 1) -> 4 monomers total, 2 monomers displaced along x-axis and 2 along y-axis, all in plane (2D)
        else:
            self.aggregate_shape = (2, 2, 1)
        
        # user might accidentally make one element of the tuple zero
        if self.aggregate_shape[0] == 0 or self.aggregate_shape[1] == 0 or self.aggregate_shape[2] == 0:
            print(" Invalid shape!  All directions must have at least 1 layer!")
            print(" Smallest possible shape (1,1,1) is a monomer - please check your input and try again")
            

        self.number_of_monomers = self.aggregate_shape[0] * self.aggregate_shape[1] * self.aggregate_shape[2]
        # allow the user to specify the displacements along each axis (x, y, z) as a vector
        if "displacement_vector" in args:
            self.displacement_vector = args["displacement_vector"]

        # default -> displacement along x = 35.47 a.u, displacement along y = 19.63 a.u., displacement along z = 8.47
        else:
            self.displacement_vector = [35.47, 19.63, 8.47]

        if "transition_dipole_moment" in args:
            self.transition_dipole_moment = args["transition_dipole_moment"]
        else:
            self.transition_dipole_moment = np.array([0, 0, 1])

        if "refractive_index" in args:
            self.refractive_index = args["refractive_index"]
        else:
            self.refractive_index = 1

        self.wvlngth_variable = np.arange(0, 650.001, 0.01)
        self.number_of_monomers = self.aggregate_shape[0] * self.aggregate_shape[1] * self.aggregate_shape[2]
    def _compute_cartesian_coordinates(self):
        """ Method to assign cartesian coordinates to the monomers
            
        Arguments
        ---------
        None

        Attributes
        ----------
        aggregate_shape : tuple
            The number of monomers along each coordinate (Nx, Ny, Nz)

        displacement_vector : numpy array of floats
            The displacement along each coordinate [Dx, Dy, Dz]

        number_of_monomers : int
            The total number of monomers

        coords : number_of_monomers x 3 numpy array of floats
            X, Y, Z coordinates for each monomer        
        """

        # get number of monomers displaced along x, y, and z directions
        _Nx = self.aggregate_shape[0]
        _Ny = self.aggregate_shape[1]
        _Nz = self.aggregate_shape[2]
        
        # initialize coords
        self.coords = np.zeros((self.number_of_monomers, 3))

        _Nc = 0
        for i in range(_Nx):
            for j in range(_Ny):
                for k in range(_Nz):
                    self.coords[_Nc, 0] = i * self.displacement_vector[0]
                    self.coords[_Nc, 1] = j * self.displacement_vector[1]
                    self.coords[_Nc, 2] = k * self.displacement_vector[2]
                    _Nc += 1
        return self.coords
    

    def _compute_H0_element(self, n, m):
        """Method to compute the matrix elements of H0

        Arguments
        ---------
        n : int
            the index of site n 
        m : int
            the index of site m 

        Returns
        -------
        H_nm : float
            The matrix elements corresponding to the interactions between sites n and m
        """
     
        H_nm = self.exciton_energy * (n == m)
        return H_nm

    def _compute_dipole_dipole_coupling(self, n, m):
        """Method to compute the dipole-dipole potential between excitons located on site n and site m

        Arguments
        ---------
        n : int
            the index of site n 
        m : int
            the index of site m 

        Attributes
        ----------
        coords : 3 x number_of_monomers numpy array of floats
            the cartesian coordinates of each monomer

        transition_dipole_moment : 1x3 numpy array of floats
            the transition dipole moment associated with the excitons

        Returns
        -------
        V_nm : float
             the dipole-dipole potential between exciton on site n and m
        """

        # calculate separation vector between site m and site n
        self.coord_array = self._compute_cartesian_coordinates()
        _r_vec = self.coord_array[n,:] - self.coord_array[m,:]

        # self.transition_dipole_moment is the transition dipole moment!
        if n != m:
            V_nm = (
                1 / (self.refractive_index**2 * np.sqrt(np.dot(_r_vec, _r_vec)) ** 3)
            ) * (
                np.dot(self.transition_dipole_moment, self.transition_dipole_moment)
                - 3
                * (
                    (
                        np.dot(self.transition_dipole_moment, _r_vec)
                        * np.dot(_r_vec, self.transition_dipole_moment)
                    )
                    / (np.sqrt(np.dot(_r_vec, _r_vec)) ** 2)
                )
            )
        else:
            V_nm = 0

        return V_nm
    

    def build_exciton_hamiltonian(self):
        """Method to build the Frenkel Exciton Hamiltonian

        Attribute
        ---------
        exciton_hamiltonian : number_of_monomers x number_of_monomers numpy array of floats
            the exciton Hamiltonian, initialized by init and to-be-filled with appropriate values
            by this function

        Notes
        -----

        """
        _N = self.number_of_monomers  # <== _N is just easier to type!
        self.exciton_hamiltonian = np.zeros((self.number_of_monomers, self.number_of_monomers))
        # nested loop to build Hamiltonian
        for _n in range(_N):
            for _m in range(_N):
                # <== call _compute_H0_element and store value -> H0
                H0 = self._compute_H0_element(_n, _m)  # <== Note self. notation
                # <== call _compute_dipole_dipole_coupling and store value -> V
                V = self._compute_dipole_dipole_coupling(
                    _n, _m
                )   # <== Note self. notation
                # <== assign H0 + V to appropriate element of self.exciton_hamiltonian
                self.exciton_hamiltonian[_n, _m] = (
                    H0 + V 
                )  # <= Note we will store the elements in hamiltonian attribute
        return self.exciton_hamiltonian
        

    """def build_2D_hamiltonian(self):
        Function that builds the Hamailtonian which models the time evolution of an excitonic system based upon the
        field free energy of the system and the dipole dipole coupling of the sysetem
         
        Attribute
        ---------
        exciton_hamiltonian : number_of_monomers x number_of_monomers numpy array of floats
            the exciton Hamiltonian, initialized by init and to-be-filled with appropriate values
            by this function
        
        
        _N = self.number_of_monomers
        exciton_hamiltonian_2D = np.zeros((_N, _N))
        for _n in range(_N):
            for _m in range(_N):
                H0 = self._compute_H0_element(_n, _m)
                V = self._compute_dipole_dipole_coupling(_n, _m)
                exciton_hamiltonian_2D[_n, _m] = (
                    H0 + V
                )
        return exciton_hamiltonian_2D
"""
    def compute_exciton_wavefunction_site_basis(self):
        """
        Will compute the single-exciton wavefunctions (approximated as Gaussians) for each site.

        ***Note:***  This is probably not the best model... better model might be a Slater function, but
        we can fix that later. Also right now the width is somewhat arbitrary.  In reality, the exciton
        has a Bohr radius that will determine the width.  We will update this later as well!

        Arguments
        ---------
        n : int
            the site index offset from python index by +1

        Attributes
        ----------
        x : 1 x _len numpy array of floats
            the spatial grid that we will evaluate the exciton wavefunction on

        x_min : float
            the minimum x-value on the grid x

        x_max : float
            the maximum x-value on the grid x

        phi : _len x number_of_monomer numpy array of floats
            the single-exciton wavefunctions for each site

        Note: self.phi[:,0] -> exciton wavefunction for site n = 1
              self.phi[:,1] -> exciton wavefunction for site n = 2
              etc.

        """
        # get distance between sites
        _dx = self.coords[1, 0] - self.coords[0, 0]

        # full-width at half-max based on distance between sites
        _fwhm = _dx / 2

        # width parameter of the Gaussian
        _c = _fwhm / 2.35482

        # normalization of Gaussian
        _a = 1 / (_c * np.sqrt(2 * np.pi))

        # create grid of x values spanning all sites in the system
        # get largest site index
        _N_max = self.number_of_monomers - 1
        # get x-value associated with largest site index
        _x_max = self.coords[_N_max, 0] + 3 * _fwhm
        # create the x-grid from 0 to _x_max
        _len = 500
        self.x = np.linspace(-_dx, _x_max, _len)

        self.phi = np.zeros((_len, self.number_of_monomers))

        for n in range(self.number_of_monomers):
            _x_n = self.coords[n, 0]
            self.phi[:, n] = np.exp(-((self.x - _x_n) ** 2) / (2 * _c**2))

        self.x_max = _x_max
        self.x_min = -_dx

    def _rk_exciton(self, dt):
        """Function that will take c(t0) and H and return c(t0 + dt)

        Arguments
        ---------
        dt : float
            the increment in time in atomic units

        Attributes
        ----------
        exciton_hamiltonian : NxN numpy array of floats
            the Hamiltonian matrix that drives the dynamics

        c_vector : 1xN numpy array of complex floats
            the current wavefunction vector that will be updated

        """
        ci = 0 + 1j
        k_1 = -ci * np.dot(self.exciton_hamiltonian, self.c_vector)
        k_2 = -ci * np.dot(self.exciton_hamiltonian, (self.c_vector + k_1 * dt / 2))
        k_3 = -ci * np.dot(self.exciton_hamiltonian, (self.c_vector + k_2 * dt / 2))
        k_4 = -ci * np.dot(self.exciton_hamiltonian, (self.c_vector + k_3 * dt))
        self.c_vector = self.c_vector + (1 / 6) * (k_1 + 2 * k_2 + 2 * k_3 + k_4) * dt
        return self.c_vector
    def _rk_exciton_density_matrix(self, dt):
        """Function that will take D and H and return D(t0 + dt)

        Arguments
        ---------
        dt : float
            the increment in time in atomic units

        Attributes
        ----------
        exciton_hamiltonian : NxN numpy array of floats
            the Hamiltonian matrix that drives the dynamics

        density_matrix : NxN numpy array of complex floats
            the current density matrix that will be updated

        """
        ci = 0 + 1j
        # going to make some temporary arrays for the partial updates of d
        # to make this more readable
        # also note I am using np.copy() here rather than setting _H = self.exciton_hamiltonian
        # see here for why: https://stackoverflow.com/questions/27538174/why-do-i-need-np-array-or-np-copy
        _H = np.copy(self.exciton_hamiltonian)
        _d0 = np.copy(self.density_matrix)

        k_1 = -ci * (np.dot(_H, _d0) - np.dot(_d0, _H))
        _d1 = _d0 + k_1 * dt / 2

        k_2 = -ci * (np.dot(_H, _d1) - np.dot(_d1, _H))
        _d2 = _d0 + k_2 * dt / 2

        k_3 = -ci * (np.dot(_H, _d2) - np.dot(_d2, _H))
        _d3 = _d0 + k_3 * dt

        k_4 = -ci * (np.dot(_H, _d3) - np.dot(_d3, _H))

        # final update - using np.copy() again
        self.density_matrix = _d0 + 1 / 6 * (k_1 + 2 * k_2 + 2 * k_3 + k_4) * dt
        
        return self.density_matrix

    def msd_psi(self, dt, N_time):
        """Method that will take dt and a number of time steps and return the mean squared displacement

        Arguments
        ---------
        dt : float
            the increment in time in atomic units
    
        N_time : int
            the number of time steps
    
         """
        prod1 = 0
        prod2 = 0
        msd_matrix = np.zeros((1, N_time))
        for i in range(self.number_of_monomers):
            for j in range(self.number_of_monomers):
                x_int = np.conj(self.c_vector[i] * self.phi[:,i]) * self.x * self.c_vector[j] * self.phi[:,j]
                x_o = np.trapz(x_int, self.x)
                for k in range(N_time):
                    self.c_temp = self._rk_exciton(dt)
                    prod1 += np.conj(self.c_temp[i] * self.phi[:,i]) * self.x ** 2 * self.c_temp[j] * self.phi[:,j]
                    prod2 += np.conj(self.c_temp[i] * self.phi[:,i]) * self.x * self.c_temp[j] * self.phi[:,j]
                    msd_matrix[0,k] = np.trapz(prod1, self.x) - 2 * x_o * np.trapz(prod2, self.x) + x_o ** 2
    
        return msd_matrix[0,:]
    
    def msd_density_matrix(self, dt, N_time):
        """Method that will take dt and a number of time steps and return the mean squared displacement

        Arguments
        ---------
        dt : float
            the increment in time in atomic units
    
        N_time : int
            the number of time steps
    
         """
        x_matrix = np.zeros((self.number_of_monomers, self.number_of_monomers))
        msd_matrix = np.zeros(N_time)

        for n in range(self.number_of_monomers):
            for m in range(self.number_of_monomers):
                x_nm = self.phi[:,m] * self.x * self.phi[:,n]
                x_matrix[n,m] = np.trapz(x_nm)
        
        x_o = np.trace(x_matrix * self.density_matrix)

        for i in range(self.number_of_monomers):
            for j in range(self.number_of_monomers):
                for k in range(N_time):
                    density_matrix = self._rk_exciton_density_matrix(dt)
                    prod1 = np.trace(density_matrix * (x_matrix ** 2))
                    prod2 = np.trace(density_matrix * x_matrix)
                    msd_matrix[k] = prod1 - 2 * x_o * prod2 * x_o ** 2
        
        return msd_matrix


    def lorentzian(self, lambda_0):
        """Helper method that will take an x array and return a lorentzian function corresponding to that x array

        Arguments
        ---------
        lambda_0 : np array of floats, 
         """
        

        return 5 ** 2 / ((self.wvlngth_variable - lambda_0) ** 2 + 5 ** 2)

    def spectrum_array(self):
        """Method that will return an array of values corresponding to a plotable spectrum

        """

        test_eigenvalues = np.linalg.eigh(self.build_exciton_hamiltonian())

        Hartree_to_J = 4.35974 * 10 ** (-18)
        h = 6.626 * 10 ** (-34)
        lightspeed = 2.998 * 10 ** 8
        m_to_nm = 10 ** 9

        eigh_J = test_eigenvalues.eigenvalues * Hartree_to_J
        eigh_wvl = m_to_nm * h * lightspeed / eigh_J

        abs_spec = np.zeros_like(self.wvlngth_variable)
        for x0 in zip(eigh_wvl):
            abs_spec += self.lorentzian(x0)
        
        return abs_spec
    
    def compute_spectrum(self):
        """method that will take values computed from spectrum_array and plot them vs wavelength
    
        """
        test_spec = self.spectrum_array()
        spectrum_plot = plt.plot(self.wvlngth_variable, test_spec, 'b-')

        return spectrum_plot 
    



"""
Unit and regression test for the wpspec package.
"""

# Import package, test suite, and other packages as needed
import wptherml
import numpy as np
import pytest
import sys

sf = wptherml.SpectrumFactory()


def test_insert_layer():
    test_1_args = {
        "wavelength_list": [501e-9, 501e-9, 1],
        "material_list": [
            "Air",
            "TiO2",
            "SiO2",
            "Air",
        ],
        "thickness_list": [0, 200e-9, 100e-9, 0],
    }

    test_2_args = {
        "wavelength_list": [501e-9, 501e-9, 1],
        "material_list": [
            "Air",
            "TiO2",
            "SiO2",
            "Ag",
            "Air",
        ],
        "thickness_list": [0, 200e-9, 100e-9, 5e-9, 0],
    }

    test1 = sf.spectrum_factory("Tmm", test_1_args)
    test2 = sf.spectrum_factory("Tmm", test_2_args)

    test1.insert_layer(3, 5e-9)
    test1.material_Ag(3)
    test1.compute_spectrum()

    assert np.allclose(test1.reflectivity_array, test2.reflectivity_array)
    assert np.allclose(test1.transmissivity_array, test2.transmissivity_array)
    assert np.allclose(test1.emissivity_array, test2.emissivity_array)


def test_remove_layer():
    test_1_args = {
        "wavelength_list": [501e-9, 501e-9, 1],
        "material_list": [
            "Air",
            "TiO2",
            "SiO2",
            "Air",
        ],
        "thickness_list": [0, 200e-9, 100e-9, 0],
    }

    test_2_args = {
        "wavelength_list": [501e-9, 501e-9, 1],
        "material_list": [
            "Air",
            "TiO2",
            "SiO2",
            "Ag",
            "Air",
        ],
        "thickness_list": [0, 200e-9, 100e-9, 5e-9, 0],
    }

    test1 = sf.spectrum_factory("Tmm", test_1_args)
    test2 = sf.spectrum_factory("Tmm", test_2_args)

    test2.remove_layer(3)
    test2.compute_spectrum()

    assert np.allclose(test1.reflectivity_array, test2.reflectivity_array)
    assert np.allclose(test1.transmissivity_array, test2.transmissivity_array)
    assert np.allclose(test1.emissivity_array, test2.emissivity_array)


def test_compute_spectrum():
    """tests public method in TmmDriver compute_spectrum()
    using several test cases with a structure as follows:
    'wavelength_list': [501e-9, 501e-9, 1],
    'material_list': ["Air", "TiO2", "SiO2", "Ag", "Au", "Pt", "AlN", "Al2O3", "Air"],
    'thickness_list': [0, 200e-9, 100e-9, 5e-9, 6e-9, 7e-9, 100e-9, 201e-9, 0]
    Under the following conditions:
    a. Normal incidence (independent of polarization)
    b. 45 degrees with s-polarilzed light
    c. 55 degrees with p-polarized light
    """

    expected_result_normal_incidence = np.array(
        [
            0.21399468,  # reflectance
            0.43566534,  # transmitttance
            0.35033999,  # emissivity
        ]
    )

    expected_result_45_degress_s_polarized = np.array(
        [
            0.09929336,  # reflectance
            0.33761702,  # transmittance
            0.56308962,  # emissivity
        ]
    )

    expected_result_55_degress_p_polarized = np.array(
        [
            0.28574605,  # reflectane
            0.37053201,  # transmittance
            0.34372194,  # emissivity
        ]
    )

    test_1_args = {
        "wavelength_list": [501e-9, 501e-9, 1],
        "material_list": [
            "Air",
            "TiO2",
            "SiO2",
            "Ag",
            "Au",
            "Pt",
            "AlN",
            "Al2O3",
            "Air",
        ],
        "thickness_list": [0, 200e-9, 100e-9, 5e-9, 6e-9, 7e-9, 100e-9, 201e-9, 0],
    }

    test_2_args = {
        "wavelength_list": [501e-9, 501e-9, 1],
        "material_list": [
            "Air",
            "TiO2",
            "SiO2",
            "Ag",
            "Au",
            "Pt",
            "AlN",
            "Al2O3",
            "Air",
        ],
        "thickness_list": [0, 200e-9, 100e-9, 5e-9, 6e-9, 7e-9, 100e-9, 201e-9, 0],
        "polarization": "s",
        "incident_angle": 45.0,
    }

    test_3_args = {
        "wavelength_list": [501e-9, 501e-9, 1],
        "material_list": [
            "Air",
            "TiO2",
            "SiO2",
            "Ag",
            "Au",
            "Pt",
            "AlN",
            "Al2O3",
            "Air",
        ],
        "thickness_list": [0, 200e-9, 100e-9, 5e-9, 6e-9, 7e-9, 100e-9, 201e-9, 0],
        "polarization": "p",
        "incident_angle": 55.0,
    }

    test1 = sf.spectrum_factory("Tmm", test_1_args)
    test2 = sf.spectrum_factory("Tmm", test_2_args)
    test3 = sf.spectrum_factory("Tmm", test_3_args)

    # test1
    assert np.isclose(
        test1.reflectivity_array[0], expected_result_normal_incidence[0], 5e-3
    )
    assert np.isclose(
        test1.transmissivity_array[0], expected_result_normal_incidence[1], 5e-3
    )
    assert np.isclose(
        test1.emissivity_array[0], expected_result_normal_incidence[2], 5e-3
    )

    # test2
    assert np.isclose(
        test2.reflectivity_array[0], expected_result_45_degress_s_polarized[0], 5e-3
    )
    assert np.isclose(
        test2.transmissivity_array[0], expected_result_45_degress_s_polarized[1], 5e-3
    )
    assert np.isclose(
        test2.emissivity_array[0], expected_result_45_degress_s_polarized[2], 5e-3
    )

    # test3
    assert np.isclose(
        test3.reflectivity_array[0], expected_result_55_degress_p_polarized[0], 5e-3
    )
    assert np.isclose(
        test3.transmissivity_array[0], expected_result_55_degress_p_polarized[1], 5e-3
    )
    assert np.isclose(
        test3.emissivity_array[0], expected_result_55_degress_p_polarized[2], 5e-3
    )


def test_compute_explicit_angle_spectrum():
    """test against a simple 230 nm Ag slab at lambda = 501 nm"""
    args = {
        "wavelength_list": [500e-9, 502e-9, 3],
        "material_list": ["Air", "Ag", "Air"],
        "thickness_list": [0, 230e-9, 0],
        "Temperature": 500,
        "therml": True,
    }

    sf = wptherml.SpectrumFactory()
    test = sf.spectrum_factory("Tmm", args)

    # hard-code in the RI values of Ag
    # 500 nm
    test._refractive_index_array[0, 1] = 0.04998114 + 3.13267845j
    # 501 nm
    test._refractive_index_array[1, 1] = 0.04997774 + 3.14242568j
    # 502 nm
    test._refractive_index_array[2, 1] = 0.04997488 + 3.15210009j

    # compute explicit-angle spectrum
    test.compute_explicit_angle_spectrum()

    # expected results from wptherml calculations with this Ag slab
    _expected_r_p = np.array(
        [
            0.98177278,
            0.98137611,
            0.97953671,
            0.97496043,
            0.96789884,
            0.96979929,
            0.991839,
        ]
    )
    _expected_r_s = np.array(
        [
            0.98180453,
            0.98219551,
            0.98388276,
            0.98740085,
            0.99207765,
            0.9964686,
            0.99930033,
        ]
    )
    _expected_t_p = np.array(
        [
            1.75624700e-08,
            1.75520517e-08,
            1.77152966e-08,
            1.92085994e-08,
            2.30502233e-08,
            1.72502686e-08,
            1.21003041e-09,
        ]
    )
    _expected_t_s = np.array(
        [
            1.75126607e-08,
            1.63066111e-08,
            1.19085368e-08,
            5.89754096e-09,
            1.89387786e-09,
            3.36368599e-10,
            1.28525918e-11,
        ]
    )
    _expected_e_p = np.array(
        [
            0.0182272,
            0.01862387,
            0.02046328,
            0.02503955,
            0.03210114,
            0.03020069,
            0.008161,
        ]
    )
    _expected_e_s = np.array(
        [
            0.01819546,
            0.01780447,
            0.01611723,
            0.01259915,
            0.00792235,
            0.0035314,
            0.00069967,
        ]
    )

    assert np.allclose(test.reflectivity_array_p[:, 1], _expected_r_p, 5e-3)
    assert np.allclose(test.reflectivity_array_s[:, 1], _expected_r_s, 5e-3)
    assert np.allclose(test.transmissivity_array_p[:, 1], _expected_t_p, 5e-3)
    assert np.allclose(test.transmissivity_array_s[:, 1], _expected_t_s, 5e-3)
    assert np.allclose(test.emissivity_array_p[:, 1], _expected_e_p, 5e-3)
    assert np.allclose(test.emissivity_array_s[:, 1], _expected_e_s, 5e-3)


def test_pm_grad():
    """
    structure = {
        'Material_List' : ['Air', 'SiO2', 'Air'],
        ### Thicknesses just chosen arbitrarily, replace with "optimal" values
        'Thickness_List': [0, 200e-9, 0],
        ### add a number to Gradient_List to optimize over more layers
        'Gradient_List': [1],
        'Lambda_List': [600e-9, 602e-9, 3],
        }
    """

    test_args = {
        "wavelength_list": [600e-9, 602e-9, 3],
        "material_list": [
            "Air",
            "SiO2",
            "Air",
        ],
        "thickness_list": [0, 200e-9, 0],
    }

    ts = sf.spectrum_factory("Tmm", test_args)
    _kz = 15475380.92450645 + 0.0j
    _phil = 3.09507618 + 0.0j
    pml = ts._compute_pm_analytical_gradient(_kz, _phil)

    print(pml)

    expected_pml = np.array(
        [
            [-719600.49694137 + 15458641.26899191j, 0 + 0j],
            [0.00000000e00 + 0j, -719600.49694137 - 15458641.26899191j],
        ]
    )

    assert np.allclose(pml, expected_pml)


def test_compute_spectrum_gradient():

    # simple structure - just glass layer
    test_args_1 = {
        "wavelength_list": [600e-9, 602e-9, 3],
        "material_list": [
            "Air",
            "SiO2",
            "Air",
        ],
        "thickness_list": [0, 200e-9, 0],
    }

    # 2-layer structure - just glass coated gold
    test_args_2 = {
        "wavelength_list": [600e-9, 602e-9, 3],
        "material_list": [
            "Air",
            "SiO2",
            "Au",
            "Air",
        ],
        "thickness_list": [0, 200e-9, 10e-9, 0],
    }

    # create instance of class
    ts1 = sf.spectrum_factory("Tmm", test_args_1)
    # hard-code RI using data from old WPTherml package
    ts1._refractive_index_array[0, :] = np.array([1 + 0j, 1.47779002 + 0.0j, 1 + 0j])
    # run compute_spectrum method so that new RI values get incorporated
    # into kz and kx arrays
    ts1.compute_spectrum()
    # compute spectrum gradient
    ts1.compute_spectrum_gradient()

    expected_T_prime_1 = 230498.36605705
    expected_R_prime_1 = -230498.36605705

    assert np.isclose(ts1.transmissivity_gradient_array[0, 0], expected_T_prime_1)
    assert np.isclose(ts1.reflectivity_gradient_array[0, 0], expected_R_prime_1)

    # create instance of class
    ts2 = sf.spectrum_factory("Tmm", test_args_2)
    # hard-code RI using data from old WPTherml package
    ts2._refractive_index_array[0, :] = np.array(
        [1 + 0j, 1.47779002 + 0.0j, 0.24463382 + 3.085112j, 1 + 0j]
    )
    # run compute_spectrum method so that new RI values get incorporated
    # into kz and kx arrays
    ts2.compute_spectrum()
    # compute spectrum gradient
    ts2.compute_spectrum_gradient()

    expected_R_prime_2 = np.array([-3782692.48735643, 32391843.05104597])
    expected_T_prime_2 = np.array([3248937.92608693, -37935758.73941045])
    expected_EPS_prime_2 = np.array([533754.5612695, 5543915.68836448])

    assert np.allclose(ts2.transmissivity_gradient_array[0, :], expected_T_prime_2)
    assert np.allclose(ts2.reflectivity_gradient_array[0, :], expected_R_prime_2)
    assert np.allclose(ts2.emissivity_gradient_array[0, :], expected_EPS_prime_2)


def test_tm_grad():
    """
    structure = {
        'Material_List' : ['Air', 'SiO2', 'Air'],
        ### Thicknesses just chosen arbitrarily, replace with "optimal" values
        'Thickness_List': [0, 200e-9, 0],
        ### add a number to Gradient_List to optimize over more layers
        'Gradient_List': [1],
        'Lambda_List': [600e-9, 602e-9, 3],
        }
    """

    test_args = {
        "wavelength_list": [600e-9, 602e-9, 3],
        "material_list": [
            "Air",
            "SiO2",
            "Air",
        ],
        "thickness_list": [0, 200e-9, 0],
    }
    # create instance of class
    ts = sf.spectrum_factory("Tmm", test_args)

    # define the z-component of wavevector exactly matching a air/sio2/air @ \lambda=600 nm
    # case from original wptherml release
    _kz = np.array(
        [10471975.51196598 + 0.0j, 15475380.92450645 + 0.0j, 10471975.51196598 + 0.0j]
    )

    # k0 = 2 * pi / \lambda
    _k0 = 10471975.511965977

    #  thickness array 200 nm glass
    _d = np.array([0, 2e-07, 0])

    # _kz * d for glass layer
    _phil = 3.09507618 + 0.0j

    # refractive index array for air/glass/air @ \lambda = 600 nm
    # exactly matching original wptherml release case
    _ri = np.array([1 + 0j, 1.47779002 + 0.0j, 1 + 0j])

    # compute the gradient of the transfer matrix
    M, theta, ctheta = ts._compute_tm_gradient(_ri, _k0, _kz, _d, 1)

    # this is the expected gradient of the transfer matrix from original wptherml release
    expected_M0 = np.array(
        [
            [-7.19600497e05 + 16652636.92188255j, -5.82076609e-11 + 6191988.90249864j],
            [0.00000000e00 - 6191988.90249864j, -7.19600497e05 - 16652636.92188255j],
        ]
    )

    assert np.allclose(M, expected_M0)

    
def test_selective_mirror_fom():
    """
    Test the computation of the selective mirror figure of merit against the following cases:

    1. The transmission exactly matches the transmissive_envelope, the reflectivity is zero everywhere
       and the weights are 0.6, 0.4 for transmission_efficiency and reflection_efficiency, respectively

    2. The transmission is zero everywhere, the reflectivity exactly matches the reflective_envelope
       and the weights are 0.25, 0.75 for transmission_efficiency and reflection_efficiency, respectively

    3. The transmission and reflectivity exactly match the windows and the weights are 0.5/0.5
    """
    # guess thickness for glass - totally arbitrary
    d1 = 500e-9

    # guess thickness for Al2O3 - totally arbitrary
    d2 = 500e-9

    # set weights for T and R for each test
    T_weight_1 = 0.6
    T_weight_2 = 0.25
    T_weight_3 = 0.5

    R_weight_1 = 0.4
    R_weight_2 = 0.75
    R_weight_3 = 0.5

    test_args1 = {
        "wavelength_list": [300e-9, 6000e-9, 1000],
        "Material_List": ["Air", "SiO2", "Al2O3", "SiO2", "Al2O3", "Air"],
        "Thickness_List": [0, d1, d2, d1, d2, 0],
        "reflective_window_wn": [2000, 2400],
        "transmissive_window_nm": [350, 700],
        "transmission_efficiency_weight": T_weight_1,
        "reflection_efficiency_weight": R_weight_1,
        "reflection_selectivity_weight" : 0,
    }
    test_args2 = {
        "wavelength_list": [300e-9, 6000e-9, 1000],
        "Material_List": ["Air", "SiO2", "Al2O3", "SiO2", "Al2O3", "Air"],
        "Thickness_List": [0, d1, d2, d1, d2, 0],
        "reflective_window_wn": [2000, 2400],
        "transmissive_window_nm": [350, 700],
        "transmission_efficiency_weight": T_weight_2,
        "reflection_efficiency_weight": R_weight_2,
        "reflection_selectivity_weight" : 0,
    }
    test_args3 = {
        "wavelength_list": [300e-9, 6000e-9, 1000],
        "Material_List": ["Air", "SiO2", "Al2O3", "SiO2", "Al2O3", "Air"],
        "Thickness_List": [0, d1, d2, d1, d2, 0],
        "reflective_window_wn": [2000, 2400],
        "transmissive_window_nm": [350, 700],
        "transmission_efficiency_weight": T_weight_3,
        "reflection_efficiency_weight": R_weight_3,
        "reflection_selectivity_weight" : 0,
    }

    # instantiate the factory
    sf = wptherml.SpectrumFactory()

    # create 3 test instances
    test1 = sf.spectrum_factory("Tmm", test_args1)
    test2 = sf.spectrum_factory("Tmm", test_args2)
    test3 = sf.spectrum_factory("Tmm", test_args3)

    # set T spectra for each test case
    test1.transmissivity_array = np.copy(test1.transmissive_envelope)  # <== perfect T
    test2.transmissivity_array = np.copy(
        np.zeros_like(test2.wavelength_array)
    )  # <== zeros everywhere
    test3.transmissivity_array = np.copy(test2.transmissive_envelope)  # <== perfect T

    # set R spectra for each test case
    test1.reflectivity_array = np.copy(
        np.zeros_like(test1.wavelength_array)
    )  # <== zeros everywhere
    test2.reflectivity_array = np.copy(test2.reflective_envelope)  # <== perfect R
    test3.reflectivity_array = np.copy(test2.reflective_envelope)  # <== perfect R

    # compute foms for each case
    test1.compute_selective_mirror_fom()
    test2.compute_selective_mirror_fom()
    test3.compute_selective_mirror_fom()

    # set the expected values
    # test 1
    _norm_1 = 1 / (T_weight_1 + R_weight_1)
    _expected_T_fom_1 = 1.0
    _expected_R_fom_1 = 0.0
    _expected_SM_fom_1 = (T_weight_1 / _norm_1) * _expected_T_fom_1 + (
        R_weight_1 / _norm_1
    ) * _expected_R_fom_1

    # test 2
    _norm_2 = 1 / (T_weight_2 + R_weight_2)
    _expected_T_fom_2 = 0.0
    _expected_R_fom_2 = 1.0
    _expected_SM_fom_2 = (T_weight_2 / _norm_2) * _expected_T_fom_2 + (
        R_weight_2 / _norm_2
    ) * _expected_R_fom_2

    # test 3
    _norm_3 = 1 / (T_weight_3 + R_weight_3)
    _expected_T_fom_3 = 1.0
    _expected_R_fom_3 = 1.0
    _expected_SM_fom_3 = (T_weight_3 / _norm_3) * _expected_T_fom_3 + (
        R_weight_3 / _norm_3
    ) * _expected_R_fom_3

    # check the weights
    assert np.isclose(test1.transmission_efficiency_weight, T_weight_1 / _norm_1)
    assert np.isclose(test2.transmission_efficiency_weight, T_weight_2 / _norm_2)
    assert np.isclose(test3.transmission_efficiency_weight, T_weight_3 / _norm_3)

    assert np.isclose(test1.reflection_efficiency_weight, R_weight_1 / _norm_1)
    assert np.isclose(test2.reflection_efficiency_weight, R_weight_2 / _norm_2)
    assert np.isclose(test3.reflection_efficiency_weight, R_weight_3 / _norm_3)

    # compare expected to computed values for all 9 foms
    assert np.isclose(test1.transmission_efficiency, _expected_T_fom_1)
    assert np.isclose(test2.transmission_efficiency, _expected_T_fom_2)
    assert np.isclose(test3.transmission_efficiency, _expected_T_fom_3)

    assert np.isclose(test1.reflection_efficiency, _expected_R_fom_1)
    assert np.isclose(test2.reflection_efficiency, _expected_R_fom_2)
    assert np.isclose(test3.reflection_efficiency, _expected_R_fom_3)

    assert np.isclose(test1.selective_mirror_fom, _expected_SM_fom_1)
    assert np.isclose(test2.selective_mirror_fom, _expected_SM_fom_2)
    assert np.isclose(test3.selective_mirror_fom, _expected_SM_fom_3)


def test_selective_mirror_fom_gradient():
    """
    Test the gradient of the selective mirror foms against finite differences for
    layer 4 for a small DBR structure
    """

    # guess thickness for glass
    d1 = 5000e-9 / (4 * 1.5)

    # guess thickness for Al2O3
    d2 = 5000e-9 / (4 * 1.85)

    test_args = {
        "wavelength_list": [300e-9, 6000e-9, 1000],
        "Material_List": [
            "Air",
            "SiO2",
            "Al2O3",
            "SiO2",
            "Al2O3",
            "SiO2",
            "Al2O3",
            "SiO2",
            "Al2O3",
            "SiO2",
            "Al2O3",
            "Air",
        ],
        "Thickness_List": [0, d1, d2, d1, d2, d1, d2, d1, d2, d1, d2, 0],
        "reflective_window_wn": [2000, 2400],
        "transmissive_window_nm": [350, 700],
    }

    sf = wptherml.SpectrumFactory()

    # create an instance of the DBR called test
    test = sf.spectrum_factory("Tmm", test_args)
    # compute FOMs
    test.compute_selective_mirror_fom()
    # compute FOM gradients
    test.compute_selective_mirror_fom_gradient()

    # select layer corresponding to the gradient element to check
    _n_layer = 4  # <== layer 4

    # set step size for finite differences
    dx = 0.01e-9

    # now do numerical gradients
    test.thickness_array[_n_layer] += dx
    test.compute_spectrum()
    test.compute_selective_mirror_fom()

    # store the foms from the forward-displaced structure
    _r_fom_f = test.reflection_efficiency
    _t_fom_f = test.transmission_efficiency

    # move to backwards-displaced structure
    test.thickness_array[_n_layer] -= 2 * dx
    test.compute_spectrum()
    test.compute_selective_mirror_fom()

    # store the foms from the backward-displaced structure
    _r_fom_b = test.reflection_efficiency
    _t_fom_b = test.transmission_efficiency

    # compute centered-finite-difference (cfd) approximation to the gradient elements for _n_layer
    _r_grad = (_r_fom_f - _r_fom_b) / (2 * dx)
    _t_grad = (_t_fom_f - _t_fom_b) / (2 * dx)

    # compare cfd to analytic gradient
    np.isclose(_r_grad, test.reflection_efficiency_gradient[_n_layer])
    np.isclose(_r_grad, test.transmission_efficiency_gradient[_n_layer])



import numpy as np
from scipy.special import spherical_jn
from scipy.special import spherical_yn
from scipy.special import jv
from scipy.special import yv
from .spectrum_driver import SpectrumDriver
from .materials import Materials


class MieDriver(SpectrumDriver, Materials):
    """Compute the absorption, scattering, and extinction spectra of a sphere using Mie theory

    Attributes
    ----------
    radius : float
        the radius of the sphere

    number_of_wavelengths : int
        the number of wavelengths over which the cross sections / efficiencies will be computed

    wavelength_array : 1 x number_of_wavelengths numpy array of floats
        the array of wavelengths in meters over which you will compute the spectra

    _size_factor_array : 1 x number_of_wavelengths numpy array of floats
        size factor of the sphere

    _relative_refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats
        the array of refractive index values corresponding to wavelength_array

    _medium_refractive_index : float
        the refractive index of the surrounding medium - assumed to be real and wavelength-independent

    q_scat : numpy array of floats
        the scattering efficiency as a function of wavelength

    q_ext : numpy array of floats
        the extenction efficiency as a function of wavelength

    q_abs : 1 x number_of_wavelengths numpy array of floats
        the absorption efficiency as a function of wavelength

    c_scat : numpy array of floats
        the scattering cross section as a function of wavelength

    c_ext : numpy array of floats
        the extinction cross section as a function of wavelength

    c_abs : 1 x number_of_wavelengths numpy array of floats
        the absorption efficiency as a function of wavelength

    _max_coefficient_n : int
        the maximum coefficient to be computed in the Mie expansion

    _n_array : 1 x _max_coefficient_n array of ints
        array of indices for the terms in the Mie expansion

    _an : _max_coefficient x number_of_wavelengths numpy array of complex floats
        the array of a coefficients in the Mie expansion

    _bn : _max_coefficientx x number_of_wavelengths numpy array of complex floats
        the array of b coefficients in the Mie expansion

    _cn : _max_coefficientx x number_of_wavelengths numpy array of complex floats
        the array of c coefficients in the Mie expansion

    _dn : _max_coefficientx x number_of_wavelengths numpy array of complex floats
        the array of d coefficients in the Mie expansion


    Returns
    -------
    None

    Examples
    --------
    >>> fill_in_with_actual_example!
    """

    def __init__(self, args):
        self.parse_input(args)
        print("Radius of the sphere is ", self.radius)
        self.ci = 0 + 1j

        self.set_refractive_indicex_array()
        self.compute_spectrum()

    def parse_input(self, args):
        if "radius" in args:
            self.radius = args["radius"]
        else:
            self.radius = 100e-9

        if "wavelength_list" in args:
            lamlist = args["wavelength_list"]
            self.wavelength_array = np.linspace(lamlist[0], lamlist[1], int(lamlist[2]))
            self.number_of_wavelengths = int(lamlist[2])
            self.wavenumber_array = 1 / self.wavelength_array
        # default wavelength array
        else:
            self.wavelength_array = np.linspace(400e-9, 800e-9, 10)
            self.number_of_wavelengths = 10
            self.wavenumber_array = 1 / self.wavelength_array

        if "sphere_material" in args:
            self.sphere_material = args["sphere_material"]
        else:
            self.sphere_material = "ag"

        if "medium_material" in args:
            self.medium_material = args["medium_material"]
        else:
            self.medium_material = "air"

        self.number_of_layers = 3
        self._refractive_index_array = np.ones(
            (self.number_of_wavelengths, 3), dtype=complex
        )
        self._relative_permeability = 1.0 + 0j
        self._size_factor_array = np.pi * 2 * self.radius / self.wavelength_array

        self.q_ext = np.zeros_like(self.wavelength_array)
        self.q_scat = np.zeros_like(self.wavelength_array)
        self.q_abs = np.zeros_like(self.wavelength_array)

    def set_refractive_indicex_array(self):
        """once materials are specified, define the refractive_index_array values"""

        # terminal layers can be air or water now
        _lmed = self.medium_material.lower()

        if _lmed == "air":
            self.material_Air(0)
        elif _lmed == "water":
            self.material_H2O(0)
        elif _lmed == "h2o":
            self.material_H2O(0)

        _lm = self.sphere_material.lower()

        # check all possible values of the material string
        # and set material as appropriate.
        # in future probably good to create a single wrapper
        # function in materials.py that will do this so
        # that MieDriver and TmmDriver can just use it rather
        # than duplicating this kind of code in both classes
        if _lm == "air":
            self.material_Air(1)
        elif _lm == "ag":
            self.material_Ag(1)
        elif _lm == "al":
            self.material_Al(1)
        elif _lm == "al2o3":
            self.material_Al2O3(1)
        elif _lm == "aln":
            self.material_AlN(1)
        elif _lm == "au":
            self.material_Au(1)
        elif _lm == "hfo2":
            self.material_HfO2(1)
        elif _lm == "pb":
            self.material_Pb(1)
        elif _lm == "polystyrene":
            self.material_polystyrene(1)
        elif _lm == "pt":
            self.material_Pt(1)
        elif _lm == "re":
            self.material_Re(1)
        elif _lm == "rh":
            self.material_Rh(1)
        elif _lm == "ru":
            self.material_Ru(1)
        elif _lm == "si":
            self.material_Si(1)
        elif _lm == "sio2":
            self.material_SiO2(1)
        elif _lm == "ta2O5":
            self.material_Ta2O5(1)
        elif _lm == "tin":
            self.material_TiN(1)
        elif _lm == "tio2":
            self.material_TiO2(1)
        elif _lm == "w":
            self.material_W(1)
        # default is SiO2
        else:
            self.material_SiO2(1)

        self._relative_refractive_index_array = (
            self._refractive_index_array[:, 1] / self._refractive_index_array[:, 0]
        )

    def compute_spectrum(self):
        """Will prepare the attributes forcomputing q_ext, q_abs, q_scat, c_abs, c_ext, c_scat
        via computing the mie coefficients

        Attributes
        ---------
        TBD


        Returns
        -------
        TBD

        """
        for i in range(0, len(self.wavelength_array)):
            # get Mie coefficients... stored in attriubutes
            m_val = self._relative_refractive_index_array[i]
            mu_val = self._relative_permeability
            x_val = self._size_factor_array[i]
            self._compute_mie_coeffients(m_val, mu_val, x_val)

            # compute q_scat
            self.q_scat[i] = self._compute_q_scattering(x_val)
            self.q_ext[i] = self._compute_q_extinction(x_val)
            self.q_abs[i] = self.q_ext[i] - self.q_scat[i]

    def _compute_s_jn(self, n, z):
        """Compute the spherical bessel function from the Bessel function
        of the first kind

        Arguments
        ---------
        n : 1 x _max_coefficient numpy array of ints
            orders of the bessel function

        z : float
            size parameter of the sphere


        Returns
        -------
        _s_jn

        Test Implemented
        ----------------
        Yes

        """
        ns = n + 0.5
        return np.sqrt(np.pi / (2 * z)) * jv(ns, z)

    def _compute_s_yn(self, n, z):
        """Compute the spherical bessel function from the Bessel function
        of the first kind

        Arguments
        ---------
        n : 1 x _max_coefficient numpy array of ints
            orders of the bessel function

        z : float
            variable passed to the bessel function


        Returns
        -------
        _s_jn

        Test Implemented
        ----------------
        Yes

        """
        ns = n + 0.5
        return np.sqrt(np.pi / (2 * z)) * yv(ns, z)

    def _compute_s_hn(self, n, z):
        """Compute the spherical bessel function h_n^{(1)}

        Arguments
        ---------
        n : 1 x _max_coefficient array of int
            orders of the bessel function

        z : float
            variable passed to the bessel function


        Returns
        -------
        _s_hn

        Test Implemented
        ----------------
        Yes
        """
        return spherical_jn(n, z) + self.ci * spherical_yn(n, z)

    def _compute_z_jn_prime(self, n, z):
        """Compute derivative of z*j_n(z) using recurrence relations

        Arguments
        ---------
        n : 1 x _max_coefficient array of int
            orders of the bessel functions
        z : float
            variable passed to the bessel function

        Returns
        -------
        _z_jn_prime

        Test Implemented
        ----------------
        Yes

        """
        return z * spherical_jn(n - 1, z) - n * spherical_jn(n, z)

    def _compute_z_hn_prime(self, n, z):
        """Compute derivative of z*h_n^{(1)}(z) using recurrence relations

        Arguments
        ---------
        n : 1 x _max_coefficient array of int
            orders of the bessel functions
        z : float
            variable passed to the bessel function

        Returns
        -------
        _z_hn_prime

        """

        return z * self._compute_s_hn(n - 1, z) - n * self._compute_s_hn(n, z)

    def _compute_mie_coeffients(self, m, mu, x):
        """computes the Mie coefficients given relative refractive index,

        Arguments
        ---------
        n : 1 x _max_coefficient array of ints
            order of the mie coefficients functions
        m : complex float
            relative refractive index of the sphere to the medium
        mu : complex float
            relative permeability of the sphere to the medium (typically 1)
        x : float
            size parameter of the sphere

        Attributes
        -------
        _an

        _bn

        _cn

        _dn

        """
        self._compute_n_array(x)
        # self._n_array will be an array from 1 to n_max

        # pre-compute terms that will be used numerous times in computing coefficients
        _jnx = spherical_jn(self._n_array, x)
        _jnmx = spherical_jn(self._n_array, m * x)
        _hnx = self._compute_s_hn(self._n_array, x)
        _xjnxp = self._compute_z_jn_prime(self._n_array, x)
        _mxjnmxp = self._compute_z_jn_prime(self._n_array, m * x)
        _xhnxp = self._compute_z_hn_prime(self._n_array, x)

        # a_n coefficients
        _a_numerator = m**2 * _jnmx * _xjnxp - mu * _jnx * _mxjnmxp
        _a_denominator = m**2 * _jnmx * _xhnxp - mu * _hnx * _mxjnmxp

        self._an = _a_numerator / _a_denominator

        # b_n coefficients
        _b_numerator = mu * _jnmx * _xjnxp - _jnx * _mxjnmxp
        _b_denominator = mu * _jnmx * _xhnxp - _hnx * _mxjnmxp

        self._bn = _b_numerator / _b_denominator

        # c_n coefficients
        _c_numerator = mu * _jnx * _xhnxp - mu * _hnx * _xjnxp
        _c_denominator = mu * _jnmx * _xhnxp - _hnx * _mxjnmxp

        self._cn = _c_numerator / _c_denominator

        # d_n coefficients
        _d_numerator = mu * m * _jnx * _xhnxp - mu * m * _hnx * _xjnxp
        _d_denominator = m**2 * _jnmx * _xhnxp - mu * _hnx * _mxjnmxp

        self._dn = _d_numerator / _d_denominator
        # return [self._an,self._bn,self._cn,self._dn]

    def _compute_q_scattering(self, x):
        """computes the scattering efficiency from the mie coefficients

        Parameters
        ----------
        m : complex float
            relative refractive index of the sphere

        mu : float
            relative permeability of the sphere

        x : float
            size parameter of the sphere, defined as 2 * pi * r / lambda
            where r is the radius of the sphere and lambda is the wavelength of illumination

        Attributes
        -------
        q_scat

        Returns
        -------
        q_scat

        """
        # c = self._compute_mie_coeffients(m, mu, x)
        an = self._an
        bn = self._bn

        q_scat = 0.0
        for i in range(0, len(an)):
            # n is i + 1
            # because i indexes the arrays, n is the multipole order
            n = i + 1
            q_scat = q_scat + 2 / x**2 * (2 * n + 1) * (
                np.abs(an[i]) ** 2 + np.abs(bn[i]) ** 2
            )

        return q_scat

    def _compute_q_extinction(self, x):
        """computes the extinction efficiency from the mie coefficients

        Parameters
        ----------

        m : complex float
            relative refractive index of the sphere

        mu : float
            relative permeability of the sphere

        x : float
            size parameter of the sphere, defined as 2 * pi * r / lambda
            where r is the radius of the sphere and lambda is the wavelength of illumination

        Attributes
        -------
        q_ext

        Returns
        -------
        q_ext

        """

        # c = self._compute_mie_coeffients(m, mu, x)
        an = self._an
        bn = self._bn

        q_ext = 0
        for i in range(0, len(an)):
            # n is i + 1
            # because i indexes the arrays, n is the multipole order
            n = i + 1
            q_ext = q_ext + 2 / x**2 * (2 * n + 1) * np.real(an[i] + bn[i])

        return q_ext

    def _compute_n_array(self, x):
        _n_max = int(x + 4 * x ** (1 / 3.0) + 2)
        self._n_array = np.copy(np.linspace(1, _n_max, _n_max, dtype=int))

    def _compute_gl(l, r, h, mu, omega_p, eps_inf, eps_d):
        """docstring goes here!"""
        zeta = h / r

        omega_l = self._compute_omega_l(l, omega_p, eps_inf, eps_d)
        # note in atomics, 4 * pi * epsilon_0 = 1
        g_l_squared = mu * omega_p / (1 * h * r**3) * (omega_l / omega_p) ** 3
        g_l_squared *= (l + 1) ** 2 / ((1 + zeta) ** (2 * l + 4)) * (1 + 1 / (2 * l))

        return np.sqrt(g_l_squared), omega_l

    def _compute_gm(l_max, r, h, mu, omega_p, eps_inf, eps_d):
        gm_array = np.zeros(len(l_max) - 2)
        omegam_array = np.zeros(len(l_max) - 2)
        for i in range(2, l_max):
            gm_array[i - 2], omegam_array[i - 2] = self._compute_gl(
                i, r, h, mu, omega_p, eps_inf, eps_d
            )

        gm = np.sqrt(np.sum(gm_array * gm_array))
        omegam = np.sum(omegam_array * gm_array * gm_array) / np.sum(
            gm_array * gm_array
        )

        return gm, omegam

    def _compute_omega_l(l, omega_p, eps_inf, eps_d):
        """docstring goes here!"""
        return omega_p / np.sqrt(eps_inf + (1 + 1 / l) * eps_d)

    def compute_hamiltonian(self, N, h, drude_dictionary, emitter_dictionary):
        """docstring goes here!"""
        # this is the wavelength relevant for the system studied in PRL 112, 253601 (2014)
        # it is not general! should think of a more general way to get the l_max value!
        # in that paper, the set omega_0 to 2.5 eV so we can convert that frequency into a wavelength
        lambda_0 = 1240 / 2.5
        # now we can get the size parameter
        x = 2 * np.pi * self.radius / lambda_0
        _l_max = int(x + 4 * x ** (1 / 3.0) + 2)

        # get parameters of the metal nanoparticle
        _omega_p = drude_dictionary["omega_p"]
        _eps_inf = drude_dictionary["eps_inf"]
        _gamma_p = drude_dictionary["gamma_p"]
        _eps_d = drude_dictionary["eps_d"]

        # get parameters of the quantum emitter
        _gamma_qe = emitter_dictionary["gamma_qe"]
        _omega_0 = emitter_dictionary["omega_0"]
        _mu = emitter_dictionary["mu"]

        # compute g_1 and omega_1
        _g1, _omega1 = self._compute_gl(
            1, self.radius, h, _mu, _omega_p, _eps_inf, _eps_d
        )

        # compute g_M and omega_M
        _gm, _omegam = self._compute_gm(
            _l_max, self.radius, h, _mu, _omega_p, _eps_inf, _eps_d
        )

        H = np.zeros((3, 3), dtype=complex)
        ci = 0 + 1j

        # diagonal elements
        H[0, 0] = _omega_0 - ci * _gamma_qe / 2
        H[1, 1] = _omega1 - ci * _gamma_p / 2
        H[2, 2] = _omegam - ci * _gamma_p / 2

        # off-diagonal for coupling to dipolar term
        H[0, 1] = _g1 * np.sqrt(N / 3)
        H[1, 0] = _g1 * np.sqrt(N / 3)

        # off-diagonal for coupling to all higher multipoles
        H[0, 2] = _gm
        H[2, 0] = _gm
